# Copyright (C) 2021, 2024, NEC corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""Unit tests for NEC Driver."""

import functools
from unittest import mock

from oslo_config import cfg
import requests
from requests import models

from cinder import context as cinder_context
from cinder import db
from cinder.db.sqlalchemy import api as sqlalchemy_api
from cinder import exception
from cinder.objects import group_snapshot as obj_group_snap
from cinder.objects import snapshot as obj_snap
from cinder.tests.unit import fake_group
from cinder.tests.unit import fake_group_snapshot
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit import test
from cinder.volume import configuration as conf
from cinder.volume import driver
from cinder.volume.drivers.hitachi import hbsd_common
from cinder.volume.drivers.hitachi import hbsd_rest
from cinder.volume.drivers.hitachi import hbsd_rest_api
from cinder.volume.drivers.nec.v import nec_v_fc
from cinder.volume import volume_types
from cinder.volume import volume_utils
from cinder.zonemanager import utils as fczm_utils

# Configuration parameter values
CONFIG_MAP = {
    'serial': '886000123456',
    'my_ip': '127.0.0.1',
    'rest_server_ip_addr': '172.16.18.108',
    'rest_server_ip_port': '23451',
    'port_id': 'CL1-A',
    'host_grp_name': 'NEC-0123456789abcdef',
    'host_mode': 'LINUX/IRIX',
    'host_wwn': '0123456789abcdef',
    'target_wwn': '1111111123456789',
    'user_id': 'user',
    'user_pass': 'password',
    'pool_name': 'test_pool',
    'auth_user': 'auth_user',
    'auth_password': 'auth_password',
}

# Dummy response for FC zoning device mapping
DEVICE_MAP = {
    'fabric_name': {
        'initiator_port_wwn_list': [CONFIG_MAP['host_wwn']],
        'target_port_wwn_list': [CONFIG_MAP['target_wwn']]}}

DEFAULT_CONNECTOR = {
    'host': 'host',
    'ip': CONFIG_MAP['my_ip'],
    'wwpns': [CONFIG_MAP['host_wwn']],
    'multipath': False,
}

CTXT = cinder_context.get_admin_context()

TEST_VOLUME = []
for i in range(5):
    volume = {}
    volume['id'] = '00000000-0000-0000-0000-{0:012d}'.format(i)
    volume['name'] = 'test-volume{0:d}'.format(i)
    volume['volume_type_id'] = '00000000-0000-0000-0000-{0:012d}'.format(i)
    if i == 3 or i == 4:
        volume['provider_location'] = None
    else:
        volume['provider_location'] = '{0:d}'.format(i)
    volume['size'] = 128
    if i == 2:
        volume['status'] = 'in-use'
    elif i == 4:
        volume['status'] = None
    else:
        volume['status'] = 'available'
    volume = fake_volume.fake_volume_obj(CTXT, **volume)
    volume.volume_type = fake_volume.fake_volume_type_obj(CTXT)
    TEST_VOLUME.append(volume)


def _volume_get(context, volume_id):
    """Return predefined volume info."""
    return TEST_VOLUME[int(volume_id.replace("-", ""))]


TEST_SNAPSHOT = []
snapshot = {}
snapshot['id'] = '10000000-0000-0000-0000-{0:012d}'.format(0)
snapshot['name'] = 'TEST_SNAPSHOT{0:d}'.format(0)
snapshot['provider_location'] = '{0:d}'.format(1)
snapshot['status'] = 'available'
snapshot['volume_id'] = '00000000-0000-0000-0000-{0:012d}'.format(0)
snapshot['volume'] = _volume_get(None, snapshot['volume_id'])
snapshot['volume_name'] = 'test-volume{0:d}'.format(0)
snapshot['volume_size'] = 128
snapshot = obj_snap.Snapshot._from_db_object(
    CTXT, obj_snap.Snapshot(),
    fake_snapshot.fake_db_snapshot(**snapshot))
TEST_SNAPSHOT.append(snapshot)

TEST_GROUP = []
for i in range(2):
    group = {}
    group['id'] = '20000000-0000-0000-0000-{0:012d}'.format(i)
    group['status'] = 'available'
    group = fake_group.fake_group_obj(CTXT, **group)
    TEST_GROUP.append(group)

TEST_GROUP_SNAP = []
group_snapshot = {}
group_snapshot['id'] = '30000000-0000-0000-0000-{0:012d}'.format(0)
group_snapshot['status'] = 'available'
group_snapshot = obj_group_snap.GroupSnapshot._from_db_object(
    CTXT, obj_group_snap.GroupSnapshot(),
    fake_group_snapshot.fake_db_group_snapshot(**group_snapshot))
TEST_GROUP_SNAP.append(group_snapshot)

# Dummy response for REST API
POST_SESSIONS_RESULT = {
    "token": "b74777a3-f9f0-4ea8-bd8f-09847fac48d3",
    "sessionId": 0,
}

GET_PORTS_RESULT = {
    "data": [
        {
            "portId": CONFIG_MAP['port_id'],
            "portType": "FIBRE",
            "portAttributes": [
                "TAR",
                "MCU",
                "RCU",
                "ELUN"
            ],
            "fabricMode": True,
            "portConnection": "PtoP",
            "lunSecuritySetting": True,
            "wwn": CONFIG_MAP['target_wwn'],
        },
    ],
}

GET_HOST_WWNS_RESULT = {
    "data": [
        {
            "hostGroupNumber": 0,
            "hostWwn": CONFIG_MAP['host_wwn'],
        },
    ],
}

COMPLETED_SUCCEEDED_RESULT = {
    "status": "Completed",
    "state": "Succeeded",
    "affectedResources": ('a/b/c/1',),
}

COMPLETED_FAILED_RESULT_LU_DEFINED = {
    "status": "Completed",
    "state": "Failed",
    "error": {
        "errorCode": {
            "SSB1": "B958",
            "SSB2": "015A",
        },
    },
}

GET_LDEV_RESULT = {
    "emulationType": "OPEN-V-CVS",
    "blockCapacity": 2097152,
    "attributes": ["CVS", "DP"],
    "status": "NML",
    "poolId": 30,
    "label": "00000000000000000000000000000000",
}

GET_LDEV_RESULT_MAPPED = {
    "emulationType": "OPEN-V-CVS",
    "blockCapacity": 2097152,
    "attributes": ["CVS", "DP"],
    "status": "NML",
    "ports": [
        {
            "portId": CONFIG_MAP['port_id'],
            "hostGroupNumber": 0,
            "hostGroupName": CONFIG_MAP['host_grp_name'],
            "lun": 1
        },
    ],
}

GET_LDEV_RESULT_SNAP = {
    "emulationType": "OPEN-V-CVS",
    "blockCapacity": 2097152,
    "attributes": ["CVS", "DP"],
    "status": "NML",
    "poolId": 30,
    "label": "10000000000000000000000000000000",
}

GET_LDEV_RESULT_PAIR = {
    "emulationType": "OPEN-V-CVS",
    "blockCapacity": 2097152,
    "attributes": ["CVS", "DP", "SS"],
    "status": "NML",
    "label": "00000000000000000000000000000000",
}

GET_LDEV_RESULT_PAIR_SNAP = {
    "emulationType": "OPEN-V-CVS",
    "blockCapacity": 2097152,
    "attributes": ["CVS", "DP", "SS"],
    "status": "NML",
    "label": "10000000000000000000000000000000",
}

GET_SNAPSHOTS_RESULT = {
    "data": [
        {
            "primaryOrSecondary": "S-VOL",
            "status": "PSUS",
            "pvolLdevId": 0,
            "muNumber": 1,
            "svolLdevId": 1,
        },
    ],
}

GET_SNAPSHOTS_RESULT_PAIR = {
    "data": [
        {
            "primaryOrSecondary": "S-VOL",
            "status": "PAIR",
            "pvolLdevId": 0,
            "muNumber": 1,
            "svolLdevId": 1,
        },
    ],
}

GET_SNAPSHOTS_RESULT_BUSY = {
    "data": [
        {
            "primaryOrSecondary": "P-VOL",
            "status": "PSUP",
            "pvolLdevId": 0,
            "muNumber": 1,
            "svolLdevId": 1,
        },
    ],
}

GET_POOLS_RESULT = {
    "data": [
        {
            "poolId": 30,
            "poolName": CONFIG_MAP['pool_name'],
            "availableVolumeCapacity": 480144,
            "totalPoolCapacity": 507780,
            "totalLocatedCapacity": 71453172,
            "virtualVolumeCapacityRate": -1,
        },
    ],
}

GET_LUNS_RESULT = {
    "data": [
        {
            "ldevId": 0,
            "lun": 1,
        },
    ],
}

GET_HOST_GROUP_RESULT = {
    "hostGroupName": CONFIG_MAP['host_grp_name'],
}

GET_HOST_GROUPS_RESULT = {
    "data": [
        {
            "hostGroupNumber": 0,
            "portId": CONFIG_MAP['port_id'],
            "hostGroupName": "NEC-test",
        },
    ],
}

GET_LDEVS_RESULT = {
    "data": [
        {
            "ldevId": 0,
            "label": "15960cc738c94c5bb4f1365be5eeed44",
        },
        {
            "ldevId": 1,
            "label": "15960cc738c94c5bb4f1365be5eeed45",
        },
    ],
}

NOTFOUND_RESULT = {
    "data": [],
}

ERROR_RESULT = {
    "errorSource": "<URL>",
    "message": "<message>",
    "solution": "<solution>",
    "messageId": "<messageId>",
    "errorCode": {
                   "SSB1": "",
                   "SSB2": "",
    }
}


def _brick_get_connector_properties(multipath=False, enforce_multipath=False):
    """Return a predefined connector object."""
    return DEFAULT_CONNECTOR


def reduce_retrying_time(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        backup_lock_waittime = hbsd_rest_api._LOCK_TIMEOUT
        backup_exec_max_waittime = hbsd_rest_api._REST_TIMEOUT
        backup_job_api_response_timeout = (
            hbsd_rest_api._JOB_API_RESPONSE_TIMEOUT)
        backup_get_api_response_timeout = (
            hbsd_rest_api._GET_API_RESPONSE_TIMEOUT)
        backup_extend_waittime = hbsd_rest_api._EXTEND_TIMEOUT
        backup_exec_retry_interval = hbsd_rest_api._EXEC_RETRY_INTERVAL
        backup_rest_server_restart_timeout = (
            hbsd_rest_api._REST_SERVER_RESTART_TIMEOUT)
        backup_state_transition_timeout = (
            hbsd_rest._STATE_TRANSITION_TIMEOUT)
        hbsd_rest_api._LOCK_TIMEOUT = 0.01
        hbsd_rest_api._REST_TIMEOUT = 0.01
        hbsd_rest_api._JOB_API_RESPONSE_TIMEOUT = 0.01
        hbsd_rest_api._GET_API_RESPONSE_TIMEOUT = 0.01
        hbsd_rest_api._EXTEND_TIMEOUT = 0.01
        hbsd_rest_api._EXEC_RETRY_INTERVAL = 0.004
        hbsd_rest_api._REST_SERVER_RESTART_TIMEOUT = 0.02
        hbsd_rest._STATE_TRANSITION_TIMEOUT = 0.01
        func(*args, **kwargs)
        hbsd_rest_api._LOCK_TIMEOUT = backup_lock_waittime
        hbsd_rest_api._REST_TIMEOUT = backup_exec_max_waittime
        hbsd_rest_api._JOB_API_RESPONSE_TIMEOUT = (
            backup_job_api_response_timeout)
        hbsd_rest_api._GET_API_RESPONSE_TIMEOUT = (
            backup_get_api_response_timeout)
        hbsd_rest_api._EXTEND_TIMEOUT = backup_extend_waittime
        hbsd_rest_api._EXEC_RETRY_INTERVAL = backup_exec_retry_interval
        hbsd_rest_api._REST_SERVER_RESTART_TIMEOUT = (
            backup_rest_server_restart_timeout)
        hbsd_rest._STATE_TRANSITION_TIMEOUT = (
            backup_state_transition_timeout)
    return wrapper


class FakeLookupService():
    """Dummy FC zoning mapping lookup service class."""

    def get_device_mapping_from_network(self, initiator_wwns, target_wwns):
        """Return predefined FC zoning mapping."""
        return DEVICE_MAP


class FakeResponse():

    def __init__(self, status_code, data=None, headers=None):
        self.status_code = status_code
        self.data = data
        self.text = data
        self.content = data
        self.headers = {'Content-Type': 'json'} if headers is None else headers

    def json(self):
        return self.data


class VStorageRESTFCDriverTest(test.TestCase):
    """Unit test class for NEC REST interface fibre channel module."""

    test_existing_ref = {'source-id': '1'}
    test_existing_ref_name = {
        'source-name': '15960cc7-38c9-4c5b-b4f1-365be5eeed45'}

    def setUp(self):
        """Set up the test environment."""
        def _set_required(opts, required):
            for opt in opts:
                opt.required = required

        # Initialize Cinder and avoid checking driver options.
        rest_required_opts = [
            opt for opt in hbsd_rest.REST_VOLUME_OPTS if opt.required]
        common_required_opts = [
            opt for opt in hbsd_common.COMMON_VOLUME_OPTS if opt.required]
        _set_required(rest_required_opts, False)
        _set_required(common_required_opts, False)
        super(VStorageRESTFCDriverTest, self).setUp()
        _set_required(rest_required_opts, True)
        _set_required(common_required_opts, True)

        self.configuration = mock.Mock(conf.Configuration)
        self.ctxt = cinder_context.get_admin_context()
        self._setup_config()
        self._setup_driver()

    def _setup_config(self):
        """Set configuration parameter values."""
        self.configuration.config_group = "REST"

        self.configuration.volume_backend_name = "RESTFC"
        self.configuration.volume_driver = (
            "cinder.volume.drivers.nec.v.nec_v_fc.VStorageFCDriver")
        self.configuration.reserved_percentage = "0"
        self.configuration.use_multipath_for_image_xfer = False
        self.configuration.enforce_multipath_for_image_xfer = False
        self.configuration.max_over_subscription_ratio = 500.0
        self.configuration.driver_ssl_cert_verify = False

        self.configuration.nec_v_storage_id = CONFIG_MAP['serial']
        self.configuration.nec_v_pools = ["30"]
        self.configuration.nec_v_snap_pool = None
        self.configuration.nec_v_ldev_range = "0-1"
        self.configuration.nec_v_target_ports = [CONFIG_MAP['port_id']]
        self.configuration.nec_v_compute_target_ports = [
            CONFIG_MAP['port_id']]
        self.configuration.nec_v_group_create = True
        self.configuration.nec_v_group_delete = True
        self.configuration.nec_v_copy_speed = 3
        self.configuration.nec_v_copy_check_interval = 3
        self.configuration.nec_v_async_copy_check_interval = 10

        self.configuration.san_login = CONFIG_MAP['user_id']
        self.configuration.san_password = CONFIG_MAP['user_pass']
        self.configuration.san_ip = CONFIG_MAP[
            'rest_server_ip_addr']
        self.configuration.san_api_port = CONFIG_MAP[
            'rest_server_ip_port']
        self.configuration.nec_v_rest_disable_io_wait = True
        self.configuration.nec_v_rest_tcp_keepalive = True
        self.configuration.nec_v_discard_zero_page = True
        self.configuration.nec_v_rest_number = "0"
        self.configuration.nec_v_lun_timeout = hbsd_rest._LUN_TIMEOUT
        self.configuration.nec_v_lun_retry_interval = (
            hbsd_rest._LUN_RETRY_INTERVAL)
        self.configuration.nec_v_restore_timeout = hbsd_rest._RESTORE_TIMEOUT
        self.configuration.nec_v_state_transition_timeout = 2
        self.configuration.nec_v_lock_timeout = hbsd_rest_api._LOCK_TIMEOUT
        self.configuration.nec_v_rest_timeout = 3
        self.configuration.nec_v_extend_timeout = (
            hbsd_rest_api._EXTEND_TIMEOUT)
        self.configuration.nec_v_exec_retry_interval = (
            hbsd_rest_api._EXEC_RETRY_INTERVAL)
        self.configuration.nec_v_rest_connect_timeout = (
            hbsd_rest_api._DEFAULT_CONNECT_TIMEOUT)
        self.configuration.nec_v_rest_job_api_response_timeout = (
            hbsd_rest_api._JOB_API_RESPONSE_TIMEOUT)
        self.configuration.nec_v_rest_get_api_response_timeout = (
            hbsd_rest_api._GET_API_RESPONSE_TIMEOUT)
        self.configuration.nec_v_rest_server_busy_timeout = (
            hbsd_rest_api._REST_SERVER_BUSY_TIMEOUT)
        self.configuration.nec_v_rest_keep_session_loop_interval = (
            hbsd_rest_api._KEEP_SESSION_LOOP_INTERVAL)
        self.configuration.nec_v_rest_another_ldev_mapped_retry_timeout = (
            hbsd_rest_api._ANOTHER_LDEV_MAPPED_RETRY_TIMEOUT)
        self.configuration.nec_v_rest_tcp_keepidle = (
            hbsd_rest_api._TCP_KEEPIDLE)
        self.configuration.nec_v_rest_tcp_keepintvl = (
            hbsd_rest_api._TCP_KEEPINTVL)
        self.configuration.nec_v_rest_tcp_keepcnt = (
            hbsd_rest_api._TCP_KEEPCNT)
        self.configuration.nec_v_host_mode_options = []

        self.configuration.nec_v_zoning_request = False

        self.configuration.san_thin_provision = True
        self.configuration.san_private_key = ''
        self.configuration.san_clustername = ''
        self.configuration.san_ssh_port = '22'
        self.configuration.san_is_local = False
        self.configuration.ssh_conn_timeout = '30'
        self.configuration.ssh_min_pool_conn = '1'
        self.configuration.ssh_max_pool_conn = '5'

        self.configuration.use_chap_auth = True
        self.configuration.chap_username = CONFIG_MAP['auth_user']
        self.configuration.chap_password = CONFIG_MAP['auth_password']

        self.configuration.safe_get = self._fake_safe_get

        CONF = cfg.CONF
        CONF.my_ip = CONFIG_MAP['my_ip']

    def _fake_safe_get(self, value):
        """Retrieve a configuration value avoiding throwing an exception."""
        try:
            val = getattr(self.configuration, value)
        except AttributeError:
            val = None
        return val

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(
        volume_utils, 'brick_get_connector_properties',
        side_effect=_brick_get_connector_properties)
    def _setup_driver(
            self, brick_get_connector_properties=None, request=None):
        """Set up the driver environment."""
        self.driver = nec_v_fc.VStorageFCDriver(
            configuration=self.configuration, db=db)
        request.side_effect = [FakeResponse(200, POST_SESSIONS_RESULT),
                               FakeResponse(200, GET_PORTS_RESULT),
                               FakeResponse(200, GET_HOST_WWNS_RESULT)]
        self.driver.do_setup(None)
        self.driver.check_for_setup_error()
        self.driver.local_path(None)
        self.driver.create_export(None, None, None)
        self.driver.ensure_export(None, None)
        self.driver.remove_export(None, None)
        self.driver.create_export_snapshot(None, None, None)
        self.driver.remove_export_snapshot(None, None)
        # stop the Loopingcall within the do_setup treatment
        self.driver.common.client.keep_session_loop.stop()

    def tearDown(self):
        self.client = None
        super(VStorageRESTFCDriverTest, self).tearDown()

    # API test cases
    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(
        volume_utils, 'brick_get_connector_properties',
        side_effect=_brick_get_connector_properties)
    def test_do_setup(self, brick_get_connector_properties, request):
        drv = nec_v_fc.VStorageFCDriver(
            configuration=self.configuration, db=db)
        self._setup_config()
        request.side_effect = [FakeResponse(200, POST_SESSIONS_RESULT),
                               FakeResponse(200, GET_PORTS_RESULT),
                               FakeResponse(200, GET_HOST_WWNS_RESULT)]
        drv.do_setup(None)
        self.assertEqual(
            {CONFIG_MAP['port_id']: CONFIG_MAP['target_wwn']},
            drv.common.storage_info['wwns'])
        self.assertEqual(1, brick_get_connector_properties.call_count)
        self.assertEqual(3, request.call_count)
        # stop the Loopingcall within the do_setup treatment
        self.driver.common.client.keep_session_loop.stop()
        self.driver.common.client.keep_session_loop.wait()

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(
        volume_utils, 'brick_get_connector_properties',
        side_effect=_brick_get_connector_properties)
    def test_do_setup_create_hg(self, brick_get_connector_properties, request):
        """Normal case: The host group not exists."""
        drv = nec_v_fc.VStorageFCDriver(
            configuration=self.configuration, db=db)
        self._setup_config()
        request.side_effect = [FakeResponse(200, POST_SESSIONS_RESULT),
                               FakeResponse(200, GET_PORTS_RESULT),
                               FakeResponse(200, NOTFOUND_RESULT),
                               FakeResponse(200, NOTFOUND_RESULT),
                               FakeResponse(200, NOTFOUND_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        drv.do_setup(None)
        self.assertEqual(
            {CONFIG_MAP['port_id']: CONFIG_MAP['target_wwn']},
            drv.common.storage_info['wwns'])
        self.assertEqual(1, brick_get_connector_properties.call_count)
        self.assertEqual(8, request.call_count)
        # stop the Loopingcall within the do_setup treatment
        self.driver.common.client.keep_session_loop.stop()
        self.driver.common.client.keep_session_loop.wait()

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(
        volume_utils, 'brick_get_connector_properties',
        side_effect=_brick_get_connector_properties)
    def test_do_setup_pool_name(self, brick_get_connector_properties, request):
        """Normal case: Specify a pool name instead of pool id"""
        drv = nec_v_fc.VStorageFCDriver(
            configuration=self.configuration)
        self._setup_config()
        tmp_pools = self.configuration.hitachi_pools
        self.configuration.hitachi_pools = [CONFIG_MAP['pool_name']]
        request.side_effect = [FakeResponse(200, POST_SESSIONS_RESULT),
                               FakeResponse(200, GET_POOLS_RESULT),
                               FakeResponse(200, GET_PORTS_RESULT),
                               FakeResponse(200, GET_HOST_WWNS_RESULT)]
        drv.do_setup(None)
        self.assertEqual(
            {CONFIG_MAP['port_id']: CONFIG_MAP['target_wwn']},
            drv.common.storage_info['wwns'])
        self.assertEqual(1, brick_get_connector_properties.call_count)
        self.assertEqual(4, request.call_count)
        self.configuration.hitachi_pools = tmp_pools
        # stop the Loopingcall within the do_setup treatment
        self.driver.common.client.keep_session_loop.stop()
        self.driver.common.client.keep_session_loop.wait()

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_create_volume(
            self, get_volume_type_qos_specs, get_volume_type_extra_specs,
            request):
        request.return_value = FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)
        get_volume_type_extra_specs.return_value = {}
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        self.driver.common._stats = {}
        self.driver.common._stats['pools'] = [
            {'location_info': {'pool_id': 30}}]
        ret = self.driver.create_volume(TEST_VOLUME[4])
        self.assertEqual('1', ret['provider_location'])
        self.assertEqual(2, request.call_count)

    @reduce_retrying_time
    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_create_volume_timeout(
            self, get_volume_type_qos_specs, get_volume_type_extra_specs,
            request):
        request.return_value = FakeResponse(
            500, ERROR_RESULT,
            headers={'Content-Type': 'json'})
        get_volume_type_extra_specs.return_value = {}
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        self.driver.common._stats = {}
        self.driver.common._stats['pools'] = [
            {'location_info': {'pool_id': 30}}]
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.create_volume,
                          TEST_VOLUME[4])
        self.assertGreater(request.call_count, 1)

    @mock.patch.object(requests.Session, "request")
    def test_delete_volume(self, request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        self.driver.delete_volume(TEST_VOLUME[0])
        self.assertEqual(4, request.call_count)

    @mock.patch.object(requests.Session, "request")
    def test_delete_volume_temporary_busy(self, request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT_PAIR),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT_BUSY),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        self.driver.delete_volume(TEST_VOLUME[0])
        self.assertEqual(7, request.call_count)

    @reduce_retrying_time
    @mock.patch.object(requests.Session, "request")
    def test_delete_volume_busy_timeout(self, request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT_PAIR),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT_BUSY),
                               FakeResponse(200, GET_LDEV_RESULT_PAIR),
                               FakeResponse(200, GET_LDEV_RESULT_PAIR),
                               FakeResponse(200, GET_LDEV_RESULT_PAIR)]
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.delete_volume,
                          TEST_VOLUME[0])
        self.assertGreater(request.call_count, 2)

    @mock.patch.object(requests.Session, "request")
    def test_extend_volume(self, request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        self.driver.extend_volume(TEST_VOLUME[0], 256)
        self.assertEqual(3, request.call_count)

    @mock.patch.object(driver.FibreChannelDriver, "get_goodness_function")
    @mock.patch.object(driver.FibreChannelDriver, "get_filter_function")
    @mock.patch.object(requests.Session, "request")
    def test__update_volume_stats(
            self, request, get_filter_function, get_goodness_function):
        request.return_value = FakeResponse(200, GET_POOLS_RESULT)
        get_filter_function.return_value = None
        get_goodness_function.return_value = None
        self.driver._update_volume_stats()
        self.assertEqual(
            'NEC', self.driver._stats['vendor_name'])
        self.assertTrue(self.driver._stats["pools"][0]['multiattach'])
        self.assertEqual(1, request.call_count)
        self.assertEqual(1, get_filter_function.call_count)
        self.assertEqual(1, get_goodness_function.call_count)

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(sqlalchemy_api, 'volume_get', side_effect=_volume_get)
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_create_snapshot(
            self, get_volume_type_qos_specs, get_volume_type_extra_specs,
            volume_get, request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        get_volume_type_extra_specs.return_value = {}
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        self.driver.common._stats = {}
        self.driver.common._stats['pools'] = [
            {'location_info': {'pool_id': 30}}]
        ret = self.driver.create_snapshot(TEST_SNAPSHOT[0])
        self.assertEqual('1', ret['provider_location'])
        self.assertEqual(5, request.call_count)

    @mock.patch.object(requests.Session, "request")
    def test_delete_snapshot(self, request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT_PAIR_SNAP),
                               FakeResponse(200, NOTFOUND_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        self.driver.delete_snapshot(TEST_SNAPSHOT[0])
        self.assertEqual(10, request.call_count)

    @mock.patch.object(requests.Session, "request")
    def test_delete_snapshot_no_pair(self, request):
        """Normal case: Delete a snapshot without pair."""
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT_SNAP),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        self.driver.delete_snapshot(TEST_SNAPSHOT[0])
        self.assertEqual(4, request.call_count)

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_create_cloned_volume(
            self, get_volume_type_qos_specs, get_volume_type_extra_specs,
            request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        get_volume_type_extra_specs.return_value = {}
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        self.driver.common._stats = {}
        self.driver.common._stats['pools'] = [
            {'location_info': {'pool_id': 30}}]
        vol = self.driver.create_cloned_volume(TEST_VOLUME[0], TEST_VOLUME[1])
        self.assertEqual('1', vol['provider_location'])
        self.assertEqual(5, request.call_count)

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_create_volume_from_snapshot(
            self, get_volume_type_qos_specs, get_volume_type_extra_specs,
            request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        get_volume_type_extra_specs.return_value = {}
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        self.driver.common._stats = {}
        self.driver.common._stats['pools'] = [
            {'location_info': {'pool_id': 30}}]
        vol = self.driver.create_volume_from_snapshot(
            TEST_VOLUME[0], TEST_SNAPSHOT[0])
        self.assertEqual('1', vol['provider_location'])
        self.assertEqual(5, request.call_count)

    @mock.patch.object(fczm_utils, "add_fc_zone")
    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    def test_initialize_connection(
            self, get_volume_type_extra_specs, request, add_fc_zone):
        self.driver.common.conf.hitachi_zoning_request = True
        self.driver.common._lookup_service = FakeLookupService()
        get_volume_type_extra_specs.return_value = {}
        request.side_effect = [FakeResponse(200, GET_HOST_WWNS_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        ret = self.driver.initialize_connection(
            TEST_VOLUME[0], DEFAULT_CONNECTOR)
        self.assertEqual('fibre_channel', ret['driver_volume_type'])
        self.assertEqual([CONFIG_MAP['target_wwn']], ret['data']['target_wwn'])
        self.assertEqual(1, ret['data']['target_lun'])
        self.assertEqual(2, request.call_count)
        self.assertEqual(1, add_fc_zone.call_count)

    @mock.patch.object(fczm_utils, "add_fc_zone")
    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    def test_initialize_connection_already_mapped(
            self, get_volume_type_extra_specs, request, add_fc_zone):
        """Normal case: ldev have already mapped."""
        self.driver.common.conf.hitachi_zoning_request = True
        self.driver.common._lookup_service = FakeLookupService()
        get_volume_type_extra_specs.return_value = {}
        request.side_effect = [
            FakeResponse(200, GET_HOST_WWNS_RESULT),
            FakeResponse(202, COMPLETED_FAILED_RESULT_LU_DEFINED),
            FakeResponse(200, GET_LUNS_RESULT),
        ]
        ret = self.driver.initialize_connection(
            TEST_VOLUME[0], DEFAULT_CONNECTOR)
        self.assertEqual('fibre_channel', ret['driver_volume_type'])
        self.assertEqual([CONFIG_MAP['target_wwn']], ret['data']['target_wwn'])
        self.assertEqual(1, ret['data']['target_lun'])
        self.assertEqual(3, request.call_count)
        self.assertEqual(1, add_fc_zone.call_count)

    @mock.patch.object(fczm_utils, "add_fc_zone")
    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    def test_initialize_connection_shared_target(
            self, get_volume_type_extra_specs, request, add_fc_zone):
        """Normal case: A target shared with other systems."""
        self.driver.common.conf.hitachi_zoning_request = True
        self.driver.common._lookup_service = FakeLookupService()
        get_volume_type_extra_specs.return_value = {}
        request.side_effect = [FakeResponse(200, NOTFOUND_RESULT),
                               FakeResponse(200, NOTFOUND_RESULT),
                               FakeResponse(200, GET_HOST_GROUPS_RESULT),
                               FakeResponse(200, GET_HOST_WWNS_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        ret = self.driver.initialize_connection(
            TEST_VOLUME[0], DEFAULT_CONNECTOR)
        self.assertEqual('fibre_channel', ret['driver_volume_type'])
        self.assertEqual([CONFIG_MAP['target_wwn']], ret['data']['target_wwn'])
        self.assertEqual(1, ret['data']['target_lun'])
        self.assertEqual(5, request.call_count)
        self.assertEqual(1, add_fc_zone.call_count)

    @mock.patch.object(fczm_utils, "remove_fc_zone")
    @mock.patch.object(requests.Session, "request")
    def test_terminate_connection(self, request, remove_fc_zone):
        self.driver.common.conf.hitachi_zoning_request = True
        self.driver.common._lookup_service = FakeLookupService()
        request.side_effect = [FakeResponse(200, GET_HOST_WWNS_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT_MAPPED),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, NOTFOUND_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        self.driver.terminate_connection(TEST_VOLUME[2], DEFAULT_CONNECTOR)
        self.assertEqual(5, request.call_count)
        self.assertEqual(1, remove_fc_zone.call_count)

    @mock.patch.object(fczm_utils, "remove_fc_zone")
    @mock.patch.object(requests.Session, "request")
    def test_terminate_connection_not_connector(self, request, remove_fc_zone):
        """Normal case: Connector is None."""
        self.driver.common.conf.hitachi_zoning_request = True
        self.driver.common._lookup_service = FakeLookupService()
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT_MAPPED),
                               FakeResponse(200, GET_HOST_GROUP_RESULT),
                               FakeResponse(200, GET_HOST_WWNS_RESULT),
                               FakeResponse(200, GET_HOST_WWNS_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT_MAPPED),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, NOTFOUND_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        self.driver.terminate_connection(TEST_VOLUME[2], None)
        self.assertEqual(8, request.call_count)
        self.assertEqual(1, remove_fc_zone.call_count)

    @mock.patch.object(fczm_utils, "remove_fc_zone")
    @mock.patch.object(requests.Session, "request")
    def test_terminate_connection_not_lun(self, request, remove_fc_zone):
        """Normal case: Lun already not exist."""
        self.driver.common.conf.hitachi_zoning_request = True
        self.driver.common._lookup_service = FakeLookupService()
        request.side_effect = [FakeResponse(200, GET_HOST_WWNS_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT)]
        self.driver.terminate_connection(TEST_VOLUME[2], DEFAULT_CONNECTOR)
        self.assertEqual(2, request.call_count)
        self.assertEqual(1, remove_fc_zone.call_count)

    @mock.patch.object(fczm_utils, "add_fc_zone")
    @mock.patch.object(requests.Session, "request")
    def test_initialize_connection_snapshot(self, request, add_fc_zone):
        self.driver.common.conf.hitachi_zoning_request = True
        self.driver.common._lookup_service = FakeLookupService()
        request.side_effect = [FakeResponse(200, GET_HOST_WWNS_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        ret = self.driver.initialize_connection_snapshot(
            TEST_SNAPSHOT[0], DEFAULT_CONNECTOR)
        self.assertEqual('fibre_channel', ret['driver_volume_type'])
        self.assertEqual([CONFIG_MAP['target_wwn']], ret['data']['target_wwn'])
        self.assertEqual(1, ret['data']['target_lun'])
        self.assertEqual(2, request.call_count)
        self.assertEqual(1, add_fc_zone.call_count)

    @mock.patch.object(fczm_utils, "remove_fc_zone")
    @mock.patch.object(requests.Session, "request")
    def test_terminate_connection_snapshot(self, request, remove_fc_zone):
        self.driver.common.conf.hitachi_zoning_request = True
        self.driver.common._lookup_service = FakeLookupService()
        request.side_effect = [FakeResponse(200, GET_HOST_WWNS_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT_MAPPED),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, NOTFOUND_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        self.driver.terminate_connection_snapshot(
            TEST_SNAPSHOT[0], DEFAULT_CONNECTOR)
        self.assertEqual(5, request.call_count)
        self.assertEqual(1, remove_fc_zone.call_count)

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_manage_existing(self, get_volume_type_qos_specs, request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_LDEVS_RESULT)]
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        ret = self.driver.manage_existing(
            TEST_VOLUME[0], self.test_existing_ref)
        self.assertEqual('1', ret['provider_location'])
        self.assertEqual(3, request.call_count)

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_manage_existing_name(self, get_volume_type_qos_specs, request):
        request.side_effect = [FakeResponse(200, GET_LDEVS_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_LDEVS_RESULT)]
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        ret = self.driver.manage_existing(
            TEST_VOLUME[0], self.test_existing_ref_name)
        self.assertEqual('1', ret['provider_location'])
        self.assertEqual(4, request.call_count)

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_manage_existing_get_size(
            self, get_volume_type_qos_specs, request):
        request.return_value = FakeResponse(200, GET_LDEV_RESULT)
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        self.driver.manage_existing_get_size(
            TEST_VOLUME[0], self.test_existing_ref)
        self.assertEqual(1, request.call_count)

    @mock.patch.object(requests.Session, "request")
    def test_manage_existing_get_size_name(self, request):
        request.side_effect = [FakeResponse(200, GET_LDEVS_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT)]
        self.driver.manage_existing_get_size(
            TEST_VOLUME[0], self.test_existing_ref_name)
        self.assertEqual(2, request.call_count)

    @mock.patch.object(requests.Session, "request")
    def test_unmanage(self, request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT)]
        self.driver.unmanage(TEST_VOLUME[0])
        self.assertEqual(2, request.call_count)

    @mock.patch.object(requests.Session, "request")
    def test_copy_image_to_volume(self, request):
        image_service = 'fake_image_service'
        image_id = 'fake_image_id'
        request.return_value = FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)
        with mock.patch.object(driver.VolumeDriver, 'copy_image_to_volume') \
                as mock_copy_image:
            self.driver.copy_image_to_volume(
                self.ctxt, TEST_VOLUME[0], image_service, image_id)
        mock_copy_image.assert_called_with(
            self.ctxt, TEST_VOLUME[0], image_service, image_id,
            disable_sparse=False)
        self.assertEqual(1, request.call_count)

    @mock.patch.object(requests.Session, "request")
    def test_update_migrated_volume(self, request):
        request.return_value = FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)
        ret = self.driver.update_migrated_volume(
            self.ctxt, TEST_VOLUME[0], TEST_VOLUME[1], "available")
        self.assertEqual(1, request.call_count)
        actual = ({'_name_id': TEST_VOLUME[1]['id'],
                   'provider_location': TEST_VOLUME[1]['provider_location']})
        self.assertEqual(actual, ret)

    def test_unmanage_snapshot(self):
        """The driver don't support unmange_snapshot."""
        self.assertRaises(
            NotImplementedError,
            self.driver.unmanage_snapshot,
            TEST_SNAPSHOT[0])

    @mock.patch.object(requests.Session, "request")
    def test_retype(self, request):
        request.return_value = FakeResponse(200, GET_LDEV_RESULT)
        new_specs = {'nec:test': 'test'}
        new_type_ref = volume_types.create(self.ctxt, 'new', new_specs)
        new_type = volume_types.get_volume_type(self.ctxt, new_type_ref['id'])
        diff = {}
        host = {
            'capabilities': {
                'location_info': {
                    'pool_id': 30,
                },
            },
        }
        ret = self.driver.retype(
            self.ctxt, TEST_VOLUME[0], new_type, diff, host)
        self.assertEqual(1, request.call_count)
        self.assertTrue(ret)

    def test_backup_use_temp_snapshot(self):
        self.assertTrue(self.driver.backup_use_temp_snapshot())

    @mock.patch.object(requests.Session, "request")
    def test_revert_to_snapshot(self, request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT_PAIR),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT)]
        self.driver.revert_to_snapshot(
            self.ctxt, TEST_VOLUME[0], TEST_SNAPSHOT[0])
        self.assertEqual(5, request.call_count)

    def test_session___call__(self):
        session = self.driver.common.client.Session('id', 'token')
        req = models.Response()
        ret = session.__call__(req)
        self.assertEqual('Session token', ret.headers['Authorization'])

    def test_create_group(self):
        ret = self.driver.create_group(self.ctxt, TEST_GROUP[0])
        self.assertIsNone(ret)

    @mock.patch.object(requests.Session, "request")
    def test_delete_group(self, request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        ret = self.driver.delete_group(
            self.ctxt, TEST_GROUP[0], [TEST_VOLUME[0]])
        self.assertEqual(4, request.call_count)
        actual = (
            {'status': TEST_GROUP[0]['status']},
            [{'id': TEST_VOLUME[0]['id'], 'status': 'deleted'}]
        )
        self.assertTupleEqual(actual, ret)

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_create_group_from_src_volume(
            self, get_volume_type_qos_specs, get_volume_type_extra_specs,
            request):
        get_volume_type_extra_specs.return_value = {}
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        self.driver.common._stats = {}
        self.driver.common._stats['pools'] = [
            {'location_info': {'pool_id': 30}}]
        ret = self.driver.create_group_from_src(
            self.ctxt, TEST_GROUP[1], [TEST_VOLUME[1]],
            source_group=TEST_GROUP[0], source_vols=[TEST_VOLUME[0]]
        )
        self.assertEqual(5, request.call_count)
        actual = (
            None, [{'id': TEST_VOLUME[1]['id'], 'provider_location': '1'}])
        self.assertTupleEqual(actual, ret)

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_create_group_from_src_snapshot(
            self, get_volume_type_qos_specs, get_volume_type_extra_specs,
            request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        get_volume_type_extra_specs.return_value = {}
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        self.driver.common._stats = {}
        self.driver.common._stats['pools'] = [
            {'location_info': {'pool_id': 30}}]
        ret = self.driver.create_group_from_src(
            self.ctxt, TEST_GROUP[0], [TEST_VOLUME[0]],
            group_snapshot=TEST_GROUP_SNAP[0], snapshots=[TEST_SNAPSHOT[0]]
        )
        self.assertEqual(5, request.call_count)
        actual = (
            None, [{'id': TEST_VOLUME[0]['id'], 'provider_location': '1'}])
        self.assertTupleEqual(actual, ret)

    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_create_group_from_src_volume_error(
            self, get_volume_type_qos_specs):
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        self.assertRaises(
            exception.VolumeDriverException, self.driver.create_group_from_src,
            self.ctxt, TEST_GROUP[1], [TEST_VOLUME[1]],
            source_group=TEST_GROUP[0], source_vols=[TEST_VOLUME[3]]
        )

    @mock.patch.object(volume_utils, 'is_group_a_cg_snapshot_type')
    def test_update_group(self, is_group_a_cg_snapshot_type):
        is_group_a_cg_snapshot_type.return_value = False
        ret = self.driver.update_group(
            self.ctxt, TEST_GROUP[0], add_volumes=[TEST_VOLUME[0]])
        self.assertTupleEqual((None, None, None), ret)

    @mock.patch.object(volume_utils, 'is_group_a_cg_snapshot_type')
    def test_update_group_error(self, is_group_a_cg_snapshot_type):
        is_group_a_cg_snapshot_type.return_value = True
        self.assertRaises(
            exception.VolumeDriverException, self.driver.update_group,
            self.ctxt, TEST_GROUP[0], add_volumes=[TEST_VOLUME[3]],
            remove_volumes=[TEST_VOLUME[0]]
        )

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(sqlalchemy_api, 'volume_get', side_effect=_volume_get)
    @mock.patch.object(volume_utils, 'is_group_a_cg_snapshot_type')
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_create_group_snapshot_non_cg(
            self, get_volume_type_qos_specs, get_volume_type_extra_specs,
            is_group_a_cg_snapshot_type, volume_get, request):
        is_group_a_cg_snapshot_type.return_value = False
        get_volume_type_extra_specs.return_value = {}
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        self.driver.common._stats = {}
        self.driver.common._stats['pools'] = [
            {'location_info': {'pool_id': 30}}]
        ret = self.driver.create_group_snapshot(
            self.ctxt, TEST_GROUP_SNAP[0], [TEST_SNAPSHOT[0]]
        )
        self.assertEqual(5, request.call_count)
        actual = (
            {'status': 'available'},
            [{'id': TEST_SNAPSHOT[0]['id'],
              'provider_location': '1',
              'status': 'available'}]
        )
        self.assertTupleEqual(actual, ret)

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(sqlalchemy_api, 'volume_get', side_effect=_volume_get)
    @mock.patch.object(volume_utils, 'is_group_a_cg_snapshot_type')
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_create_group_snapshot_cg(
            self, get_volume_type_qos_specs, get_volume_type_extra_specs,
            is_group_a_cg_snapshot_type, volume_get, request):
        is_group_a_cg_snapshot_type.return_value = True
        get_volume_type_extra_specs.return_value = {}
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        request.side_effect = [FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT_PAIR),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT)]
        self.driver.common._stats = {}
        self.driver.common._stats['pools'] = [
            {'location_info': {'pool_id': 30}}]
        ret = self.driver.create_group_snapshot(
            self.ctxt, TEST_GROUP_SNAP[0], [TEST_SNAPSHOT[0]]
        )
        self.assertEqual(6, request.call_count)
        actual = (
            None,
            [{'id': TEST_SNAPSHOT[0]['id'],
              'provider_location': '1',
              'status': 'available'}]
        )
        self.assertTupleEqual(actual, ret)

    @mock.patch.object(requests.Session, "request")
    def test_delete_group_snapshot(self, request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT_PAIR_SNAP),
                               FakeResponse(200, NOTFOUND_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        ret = self.driver.delete_group_snapshot(
            self.ctxt, TEST_GROUP_SNAP[0], [TEST_SNAPSHOT[0]])
        self.assertEqual(10, request.call_count)
        actual = (
            {'status': TEST_GROUP_SNAP[0]['status']},
            [{'id': TEST_SNAPSHOT[0]['id'], 'status': 'deleted'}]
        )
        self.assertTupleEqual(actual, ret)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt

from cinder import exception
from cinder.tests.unit.privsep.targets import fake_nvmet_lib
from cinder.tests.unit.targets import targets_fixture as tf
from cinder import utils
from cinder.volume.targets import nvmet
# This must go after fake_nvmet_lib has been imported (thus the noqa)
from cinder.privsep.targets import nvmet as priv_nvmet  # noqa


@ddt.ddt
class TestNVMETDriver(tf.TargetDriverFixture):

    def setUp(self):
        super(TestNVMETDriver, self).setUp()

        self.configuration.target_prefix = 'nvme-subsystem-1'
        self.configuration.target_protocol = 'nvmet_rdma'
        self.target = nvmet.NVMET(root_helper=utils.get_root_helper(),
                                  configuration=self.configuration)
        self.target.share_targets = False
        fake_nvmet_lib.reset_mock()

    def test_supports_shared(self):
        self.assertTrue(self.target.SHARED_TARGET_SUPPORT)

    @mock.patch.object(nvmet.nvmeof.NVMeOF, 'initialize_connection')
    @mock.patch.object(nvmet.NVMET, '_map_volume')
    def test_initialize_connection_non_shared(self, mock_map, mock_init_conn):
        """Non shared initialize doesn't do anything (calls NVMeOF)."""
        res = self.target.initialize_connection(mock.sentinel.volume,
                                                mock.sentinel.connector)
        self.assertEqual(mock_init_conn.return_value, res)
        mock_init_conn.assert_called_once_with(mock.sentinel.volume,
                                               mock.sentinel.connector)
        mock_map.assert_not_called()

    @mock.patch.object(nvmet.NVMET, '_get_nvme_uuid')
    @mock.patch('os.path.exists')
    @mock.patch.object(nvmet.NVMET, '_get_connection_properties')
    @mock.patch.object(nvmet.nvmeof.NVMeOF, 'initialize_connection')
    @mock.patch.object(nvmet.NVMET, '_map_volume')
    def test_initialize_connection_shared(
            self, mock_map, mock_init_conn, mock_get_conn_props, mock_exists,
            mock_uuid):
        """When sharing, the initialization maps the volume."""
        self.mock_object(self.target, 'share_targets', True)
        mock_map.return_value = (mock.sentinel.nqn, mock.sentinel.nsid)
        vol = mock.Mock()
        res = self.target.initialize_connection(vol, mock.sentinel.connector)

        expected = {'driver_volume_type': 'nvmeof',
                    'data': mock_get_conn_props.return_value}
        self.assertEqual(expected, res)

        mock_init_conn.assert_not_called()
        mock_exists.assert_called_once_with(vol.provider_location)
        mock_map.assert_called_once_with(vol,
                                         vol.provider_location,
                                         mock.sentinel.connector)
        mock_uuid.assert_called_once_with(vol)
        mock_get_conn_props.assert_called_once_with(
            mock.sentinel.nqn,
            self.target.target_ips,
            self.target.target_port,
            self.target.nvme_transport_type,
            mock.sentinel.nsid,
            mock_uuid.return_value)

    @mock.patch.object(nvmet.NVMET, '_get_nvme_uuid')
    @mock.patch('os.path.exists', return_value=False)
    @mock.patch.object(nvmet.NVMET, '_get_connection_properties')
    @mock.patch.object(nvmet.nvmeof.NVMeOF, 'initialize_connection')
    @mock.patch.object(nvmet.NVMET, '_map_volume')
    def test_initialize_connection_shared_no_path(
            self, mock_map, mock_init_conn, mock_get_conn_props, mock_exists,
            mock_uuid):
        """Fails if the provided path is not present in the system."""
        self.mock_object(self.target, 'share_targets', True)
        mock_map.return_value = (mock.sentinel.nqn, mock.sentinel.nsid)
        vol = mock.Mock()
        self.assertRaises(exception.InvalidConfigurationValue,
                          self.target.initialize_connection,
                          vol, mock.sentinel.connector)

        mock_init_conn.assert_not_called()
        mock_exists.assert_called_once_with(vol.provider_location)
        mock_map.assert_not_called()
        mock_uuid.assert_not_called()
        mock_get_conn_props.assert_not_called()

    @mock.patch.object(nvmet.NVMET, 'get_nvmeof_location')
    @mock.patch.object(nvmet.NVMET, '_map_volume')
    def test_create_export(self, mock_map, mock_location):
        """When not sharing, the export maps the volume."""
        mock_map.return_value = (mock.sentinel.nqn, mock.sentinel.nsid)

        res = self.target.create_export(mock.sentinel.context,
                                        mock.sentinel.vol,
                                        mock.sentinel.volume_path)

        self.assertEqual({'location': mock_location.return_value, 'auth': ''},
                         res)
        mock_map.assert_called_once_with(mock.sentinel.vol,
                                         mock.sentinel.volume_path)
        mock_location.assert_called_once_with(mock.sentinel.nqn,
                                              self.target.target_ips,
                                              self.target.target_port,
                                              self.target.nvme_transport_type,
                                              mock.sentinel.nsid)

    @mock.patch.object(nvmet.NVMET, 'get_nvmeof_location')
    @mock.patch.object(nvmet.NVMET, '_map_volume')
    def test_create_export_shared(self, mock_map, mock_location):
        """When sharing, the export just stores the volume path as location."""
        self.mock_object(self.target, 'share_targets', True)

        res = self.target.create_export(mock.sentinel.context,
                                        mock.sentinel.vol,
                                        mock.sentinel.volume_path)

        self.assertEqual({'location': mock.sentinel.volume_path, 'auth': ''},
                         res)
        mock_map.assert_not_called()
        mock_location.assert_not_called()

    @mock.patch('oslo_concurrency.lockutils.lock')
    @mock.patch.object(nvmet.NVMET, '_get_nvme_uuid')
    @mock.patch.object(nvmet.NVMET, '_ensure_port_exports')
    @mock.patch.object(nvmet.NVMET, '_ensure_subsystem_exists')
    @mock.patch.object(nvmet.NVMET, '_get_target_nqn')
    def test__map_volume(self, mock_nqn, mock_subsys, mock_port, mock_uuid,
                         mock_lock):
        """Normal volume mapping."""
        vol = mock.Mock()
        res = self.target._map_volume(vol, mock.sentinel.volume_path,
                                      mock.sentinel.connector)

        expected = (mock_nqn.return_value, mock_subsys.return_value)
        self.assertEqual(res, expected)

        mock_nqn.assert_called_once_with(vol.id, mock.sentinel.connector)
        mock_uuid.assert_called_once_with(vol)
        mock_subsys.assert_called_once_with(mock_nqn.return_value,
                                            mock.sentinel.volume_path,
                                            mock_uuid.return_value)
        mock_port.assert_called_once_with(mock_nqn.return_value,
                                          self.target.target_ips,
                                          self.target.target_port,
                                          self.target.nvme_transport_type,
                                          self.target.nvmet_port_id)
        mock_lock.assert_called()

    @ddt.data((ValueError, None), (None, IndexError))
    @ddt.unpack
    @mock.patch('oslo_concurrency.lockutils.lock')
    @mock.patch.object(nvmet.NVMET, '_get_nvme_uuid')
    @mock.patch.object(nvmet.NVMET, '_ensure_port_exports')
    @mock.patch.object(nvmet.NVMET, '_ensure_subsystem_exists')
    @mock.patch.object(nvmet.NVMET, '_get_target_nqn')
    def test__map_volume_error(self, subsys_effect, port_effect, mock_nqn,
                               mock_subsys, mock_port, mock_uuid, mock_lock):
        """Failing create target executing subsystem or port creation."""
        mock_subsys.side_effect = subsys_effect
        mock_port.side_effect = port_effect
        mock_nqn.return_value = mock.sentinel.nqn
        mock_uuid.return_value = mock.sentinel.uuid
        vol = mock.Mock()

        self.assertRaises(nvmet.NVMETTargetAddError,
                          self.target._map_volume,
                          vol,
                          mock.sentinel.volume_path,
                          mock.sentinel.connector)

        mock_nqn.assert_called_once_with(vol.id, mock.sentinel.connector)
        mock_uuid.assert_called_once_with(vol)
        mock_subsys.assert_called_once_with(mock.sentinel.nqn,
                                            mock.sentinel.volume_path,
                                            mock.sentinel.uuid)
        if subsys_effect:
            mock_port.assert_not_called()
        else:
            mock_port.assert_called_once_with(mock.sentinel.nqn,
                                              self.target.target_ips,
                                              self.target.target_port,
                                              self.target.nvme_transport_type,
                                              self.target.nvmet_port_id)
        mock_lock.assert_called()

    @mock.patch.object(nvmet.NVMET, '_ensure_namespace_exists')
    @mock.patch.object(priv_nvmet, 'Subsystem')
    def test__ensure_subsystem_exists_already_exists(self, mock_subsys,
                                                     mock_namespace):
        """Skip subsystem creation if already exists."""
        nqn = 'nqn.nvme-subsystem-1-uuid'
        res = self.target._ensure_subsystem_exists(nqn,
                                                   mock.sentinel.vol_path,
                                                   mock.sentinel.uuid)
        self.assertEqual(mock_namespace.return_value, res)
        mock_subsys.assert_called_once_with(nqn)
        mock_subsys.setup.assert_not_called()
        mock_namespace.assert_called_once_with(mock_subsys.return_value,
                                               mock.sentinel.vol_path,
                                               mock.sentinel.uuid)

    @mock.patch.object(nvmet.NVMET, '_ensure_namespace_exists')
    @mock.patch('oslo_utils.uuidutils.generate_uuid')
    @mock.patch.object(priv_nvmet, 'Subsystem')
    def test__ensure_subsystem_exists(self, mock_subsys, mock_uuid,
                                      mock_namespace):
        """Create subsystem when it doesn't exist."""
        mock_subsys.side_effect = priv_nvmet.NotFound
        mock_uuid.return_value = 'uuid'
        nqn = 'nqn.nvme-subsystem-1-uuid'
        self.target._ensure_subsystem_exists(nqn,
                                             mock.sentinel.vol_path,
                                             mock.sentinel.uuid)
        mock_subsys.assert_called_once_with(nqn)
        expected_section = {
            'allowed_hosts': [],
            'attr': {'allow_any_host': '1'},
            'namespaces': [{'device': {'nguid': 'uuid',
                                       'uuid': mock.sentinel.uuid,
                                       'path': mock.sentinel.vol_path},
                            'enable': 1,
                            'nsid': self.target.nvmet_ns_id}],
            'nqn': nqn
        }
        mock_subsys.setup.assert_called_once_with(expected_section)
        mock_namespace.assert_not_called()

    @mock.patch('oslo_utils.uuidutils.generate_uuid')
    def test__namespace_dict(self, mock_uuid):
        """For not shared nguid is randomly generated."""
        res = self.target._namespace_dict(mock.sentinel.uuid,
                                          mock.sentinel.volume_path,
                                          mock.sentinel.ns_id)
        expected = {"device": {"nguid": str(mock_uuid.return_value),
                               "uuid": mock.sentinel.uuid,
                               "path": mock.sentinel.volume_path},
                    "enable": 1,
                    "nsid": mock.sentinel.ns_id}
        self.assertEqual(expected, res)
        mock_uuid.assert_called_once()

    @mock.patch('oslo_utils.uuidutils.generate_uuid')
    def test__namespace_dict_shared(self, mock_uuid):
        """For shared uuid = nguid."""
        self.mock_object(self.target, 'share_targets', True)
        res = self.target._namespace_dict(mock.sentinel.uuid,
                                          mock.sentinel.volume_path,
                                          mock.sentinel.ns_id)
        expected = {"device": {"nguid": mock.sentinel.uuid,
                               "uuid": mock.sentinel.uuid,
                               "path": mock.sentinel.volume_path},
                    "enable": 1,
                    "nsid": mock.sentinel.ns_id}
        self.assertEqual(expected, res)
        mock_uuid.assert_not_called

    def test__ensure_namespace_exist_exists(self):
        """Nothing to do if the namespace is already mapped."""
        base_path = '/dev/stack-volumes-lvmdriver-1/volume-'
        volume_path = f'{base_path}uuid2'
        subsys = mock.Mock()
        ns_other = mock.Mock(**{'get_attr.return_value': f'{base_path}uuid1'})
        ns_found = mock.Mock(**{'get_attr.return_value': volume_path})
        # nw_other appears twice to confirm we stop when found
        subsys.namespaces = [ns_other, ns_found, ns_other]
        res = self.target._ensure_namespace_exists(subsys, volume_path,
                                                   mock.sentinel.uuid)
        self.assertEqual(ns_found.nsid, res)
        ns_other.get_attr.assert_called_once_with('device', 'path')
        ns_found.get_attr.assert_called_once_with('device', 'path')

    @mock.patch.object(priv_nvmet, 'Namespace')
    @mock.patch.object(nvmet.NVMET, '_namespace_dict')
    @mock.patch.object(nvmet.NVMET, '_get_available_namespace_id')
    def test__ensure_namespace_exist_create(self, mock_get_nsid, mock_ns_dict,
                                            mock_ns):
        """Create the namespace when the path is not mapped yet."""
        base_path = '/dev/stack-volumes-lvmdriver-1/volume-'
        subsys = mock.Mock()
        ns_other = mock.Mock(**{'get_attr.return_value': f'{base_path}uuid1'})
        subsys.namespaces = [ns_other]
        res = self.target._ensure_namespace_exists(subsys,
                                                   mock.sentinel.volume_path,
                                                   mock.sentinel.uuid)
        self.assertEqual(mock_get_nsid.return_value, res)
        ns_other.get_attr.assert_called_once_with('device', 'path')
        mock_get_nsid.assert_called_once_with(subsys)
        mock_ns_dict.assert_called_once_with(mock.sentinel.uuid,
                                             mock.sentinel.volume_path,
                                             mock_get_nsid.return_value)
        mock_ns.setup.assert_called_once_with(subsys,
                                              mock_ns_dict.return_value)

    def test__get_available_namespace_id(self):
        """For non shared we always return the value from the config."""
        res = self.target._get_available_namespace_id(mock.Mock())
        self.assertEqual(self.target.nvmet_ns_id, res)

    def test__get_available_namespace_id_none_used(self):
        """For shared, on empty subsystem return the configured value."""
        self.mock_object(self.target, 'share_targets', True)
        subsys = mock.Mock(namespaces=[])
        res = self.target._get_available_namespace_id(subsys)
        self.assertEqual(self.target.nvmet_ns_id, res)

    def test__get_available_namespace_id_no_gaps(self):
        """For shared, if there are no gaps in ids return next."""
        self.mock_object(self.target, 'share_targets', True)
        expected = self.target.nvmet_ns_id + 2
        subsys = mock.Mock(namespaces=[mock.Mock(nsid=expected - 1),
                                       mock.Mock(nsid=expected - 2)])
        res = self.target._get_available_namespace_id(subsys)
        self.assertEqual(expected, res)

    def test__get_available_namespace_id_gap_value(self):
        """For shared, if there is a gap any of them is valid."""
        self.mock_object(self.target, 'share_targets', True)
        lower = self.target.nvmet_ns_id
        subsys = mock.Mock(namespaces=[mock.Mock(nsid=lower + 3),
                                       mock.Mock(nsid=lower)])
        res = self.target._get_available_namespace_id(subsys)
        self.assertTrue(res in [lower + 2, lower + 1])

    @mock.patch.object(priv_nvmet, 'Port')
    def test__ensure_port_exports_already_does(self, mock_port):
        """Skips port creation and subsystem export since they both exist."""
        nqn = 'nqn.nvme-subsystem-1-uuid'
        port_id = 1
        mock_port.return_value.subsystems = [nqn]
        self.target._ensure_port_exports(nqn,
                                         [mock.sentinel.addr],
                                         mock.sentinel.port,
                                         mock.sentinel.transport,
                                         port_id)
        mock_port.assert_called_once_with(port_id)
        mock_port.setup.assert_not_called()
        mock_port.return_value.add_subsystem.assert_not_called()

    @mock.patch.object(priv_nvmet, 'Port')
    def test__ensure_port_exports_port_exists_not_exported(self, mock_port):
        """Skips port creation if exists but exports subsystem."""
        nqn = 'nqn.nvme-subsystem-1-vol-2-uuid'
        port_id = 1
        mock_port.return_value.subsystems = ['nqn.nvme-subsystem-1-vol-1-uuid']
        self.target._ensure_port_exports(nqn,
                                         [mock.sentinel.addr],
                                         mock.sentinel.port,
                                         mock.sentinel.transport,
                                         port_id)
        mock_port.assert_called_once_with(port_id)
        mock_port.setup.assert_not_called()
        mock_port.return_value.add_subsystem.assert_called_once_with(nqn)

    @mock.patch.object(priv_nvmet, 'Port')
    def test__ensure_port_exports_port(self, mock_port):
        """Creates the port and export the subsystem when they don't exist."""
        nqn = 'nqn.nvme-subsystem-1-vol-2-uuid'
        port_id = 1
        mock_port.side_effect = priv_nvmet.NotFound
        self.target._ensure_port_exports(nqn,
                                         [mock.sentinel.addr,
                                          mock.sentinel.addr2],
                                         mock.sentinel.port,
                                         mock.sentinel.transport,
                                         port_id)
        new_port1 = {'addr': {'adrfam': 'ipv4',
                              'traddr': mock.sentinel.addr,
                              'treq': 'not specified',
                              'trsvcid': mock.sentinel.port,
                              'trtype': mock.sentinel.transport},
                     'portid': port_id,
                     'referrals': [],
                     'subsystems': [nqn]}
        new_port2 = new_port1.copy()
        new_port2['portid'] = port_id + 1
        new_port2['addr'] = new_port1['addr'].copy()
        new_port2['addr']['traddr'] = mock.sentinel.addr2

        self.assertEqual(2, mock_port.call_count)
        self.assertEqual(2, mock_port.setup.call_count)
        mock_port.assert_has_calls([
            mock.call(port_id),
            mock.call.setup(self.target._nvmet_root, new_port1),
            mock.call(port_id + 1),
            mock.call.setup(self.target._nvmet_root, new_port2)
        ])
        mock_port.return_value.assert_not_called()

    @mock.patch.object(nvmet.NVMET, '_locked_unmap_volume')
    def test_terminate_connection(self, mock_unmap):
        """For non shared there's nothing to do."""
        self.target.terminate_connection(mock.sentinel.vol,
                                         mock.sentinel.connector)
        mock_unmap.assert_not_called()

    @mock.patch.object(nvmet.NVMET, '_locked_unmap_volume')
    def test_terminate_connection_shared(self, mock_unmap):
        """For shared the volume must be unmapped."""
        self.mock_object(self.target, 'share_targets', True)
        vol = mock.Mock()
        self.target.terminate_connection(vol,
                                         mock.sentinel.connector)
        mock_unmap.assert_called_once_with(vol,
                                           mock.sentinel.connector)

    @mock.patch.object(nvmet.NVMET, '_locked_unmap_volume')
    def test_remove_export(self, mock_unmap):
        """For non shared the volume must be unmapped."""
        vol = mock.Mock()
        self.target.remove_export(mock.sentinel.context,
                                  vol)
        mock_unmap.assert_called_once_with(vol)

    @mock.patch.object(nvmet.NVMET, '_locked_unmap_volume')
    def test_remove_export_shared(self, mock_unmap):
        """For shared there's nothing to do."""
        self.mock_object(self.target, 'share_targets', True)
        self.target.remove_export(mock.sentinel.context,
                                  mock.sentinel.vol)
        mock_unmap.assert_not_called()

    @mock.patch('oslo_concurrency.lockutils.lock')
    @mock.patch.object(nvmet.NVMET, '_get_nqns_for_location', return_value=[])
    @mock.patch.object(nvmet.NVMET, '_get_target_nqn')
    @mock.patch.object(nvmet.NVMET, '_unmap_volume')
    def test__locked_unmap_volume_no_nqn(self, mock_unmap, mock_nqn, mock_nqns,
                                         mock_lock):
        """Nothing to do with no subsystem when sharing and no connector."""
        self.mock_object(self.target, 'share_targets', True)

        vol = mock.Mock()
        self.target._locked_unmap_volume(vol, connector=None)

        mock_lock.assert_called()
        mock_nqn.assert_not_called()
        mock_nqns.assert_called_once_with(vol.provider_location)
        mock_unmap.assert_not_called()

    @mock.patch('oslo_concurrency.lockutils.lock')
    @mock.patch.object(nvmet.NVMET, '_get_nqns_for_location')
    @mock.patch.object(nvmet.NVMET, '_get_target_nqn')
    @mock.patch.object(nvmet.NVMET, '_unmap_volume')
    def test__locked_unmap_volume_non_shared(self, mock_unmap, mock_nqn,
                                             mock_nqns, mock_lock):
        """Unmap locked with non sharing and no connector."""
        vol = mock.Mock()
        self.target._locked_unmap_volume(vol, connector=None)

        mock_lock.assert_called()
        mock_nqn.assert_called_once_with(vol.id, None)
        mock_nqns.assert_not_called()
        mock_unmap.assert_called_once_with(vol, mock_nqn.return_value)

    @mock.patch('oslo_concurrency.lockutils.lock')
    @mock.patch.object(nvmet.NVMET, '_get_nqns_for_location')
    @mock.patch.object(nvmet.NVMET, '_get_target_nqn')
    @mock.patch.object(nvmet.NVMET, '_unmap_volume')
    def test__locked_unmap_volume_shared_multiple(self, mock_unmap, mock_nqn,
                                                  mock_nqns, mock_lock):
        """Unmap locked with sharing and no connector, having multiple nqns."""
        self.mock_object(self.target, 'share_targets', True)
        vol = mock.Mock()
        mock_nqns.return_value = [mock.sentinel.nqn1, mock.sentinel.nqn2]

        self.target._locked_unmap_volume(vol, connector=None)

        mock_lock.assert_called()
        mock_nqn.assert_not_called()
        mock_nqns.assert_called_once_with(vol.provider_location)

        expected = [mock.call(vol, mock.sentinel.nqn1),
                    mock.call(vol, mock.sentinel.nqn2)]
        mock_unmap.assert_has_calls(expected)
        self.assertEqual(2, mock_unmap.call_count)

    @mock.patch.object(nvmet.NVMET, '_get_target_nqn')
    @mock.patch.object(priv_nvmet, 'Subsystem')
    def test__unmap_volume_no_subsys(self, mock_subsys, mock_nqn):
        """Nothing to do it there is no subsystem."""
        mock_subsys.side_effect = priv_nvmet.NotFound
        vol = mock.Mock()
        # This port is used just to confirm we don't reach that part
        port = mock.Mock(subsystems=[mock.sentinel.port])
        self.mock_object(priv_nvmet.Root, 'ports', [port])

        self.target._unmap_volume(vol, mock.sentinel.nqn)
        mock_subsys.assert_called_once_with(mock.sentinel.nqn)

        port.remove_subsystem.assert_not_called()

    @mock.patch.object(priv_nvmet, 'Subsystem')
    def test__unmap_volume_not_shared(self, mock_subsys):
        """Non shared assumes the subsystem is empty."""
        vol = mock.Mock()
        # The ns is used to confirm we don't check it
        ns = mock.Mock(**{'get_attr.return_value': vol.provider_location})
        subsys = mock_subsys.return_value
        subsys.nqn = mock.sentinel.nqn
        subsys.namespaces = [ns]

        port = mock.Mock(subsystems=[subsys.nqn])
        self.mock_object(priv_nvmet.Root, 'ports', [port])

        self.target._unmap_volume(vol, mock.sentinel.nqn)

        mock_subsys.assert_called_once_with(mock.sentinel.nqn)

        ns.get_attr.assert_not_called()
        ns.delete.assert_not_called()

        port.remove_subsystem.assert_called_once_with(mock.sentinel.nqn)
        subsys.delete.assert_called_once_with()

    @mock.patch.object(priv_nvmet, 'Subsystem')
    def test__unmap_volume_shared_more_ns(self, mock_subsys):
        """For shared don't unexport subsys if there are other ns."""
        self.mock_object(self.target, 'share_targets', True)
        vol = mock.Mock()

        ns = mock.Mock(**{'get_attr.return_value': vol.provider_location})
        subsys = mock_subsys.return_value
        subsys.namespaces = [ns]

        # Use this port to confirm we don't reach that point
        port = mock.Mock(subsystems=[subsys])
        self.mock_object(priv_nvmet.Root, 'ports', [port])

        self.target._unmap_volume(vol, mock.sentinel.nqn)

        mock_subsys.assert_called_once_with(mock.sentinel.nqn)

        ns.get_attr.assert_called_once_with('device', 'path')
        ns.delete.assert_called_once_with()

        port.remove_subsystem.assert_not_called()
        mock_subsys.return_value.delete.assert_not_called()

    @mock.patch('oslo_concurrency.lockutils.lock')
    @mock.patch.object(nvmet.NVMET, '_get_target_nqn')
    @mock.patch.object(priv_nvmet, 'Subsystem')
    def test__unmap_volume_shared_last_ns(self, mock_subsys, mock_nqn,
                                          mock_lock):
        """For shared unexport subsys if there are no other ns."""
        self.mock_object(self.target, 'share_targets', True)
        vol = mock.Mock()

        ns = mock.Mock(**{'get_attr.return_value': vol.provider_location})
        nss = [ns]
        ns.delete.side_effect = nss.clear
        subsys = mock_subsys.return_value
        subsys.nqn = mock.sentinel.nqn
        subsys.namespaces = nss

        port = mock.Mock(subsystems=[subsys.nqn])
        self.mock_object(priv_nvmet.Root, 'ports', [port])

        self.target._unmap_volume(vol, mock.sentinel.nqn)

        mock_subsys.assert_called_once_with(mock.sentinel.nqn)

        ns.get_attr.assert_called_once_with('device', 'path')
        ns.delete.assert_called_once_with()

        port.remove_subsystem.assert_called_once_with(mock.sentinel.nqn)
        mock_subsys.return_value.delete.assert_called_once_with()

    def test__get_target_nqn(self):
        """Non shared uses volume id for subsystem name."""
        res = self.target._get_target_nqn('volume_id', None)
        self.assertEqual('nqn.nvme-subsystem-1-volume_id', res)

    def test__get_target_nqn_shared(self):
        """Shared uses connector's hostname for subsystem name."""
        self.mock_object(self.target, 'share_targets', True)
        res = self.target._get_target_nqn('volume_id', {'host': 'localhost'})
        self.assertEqual('nqn.nvme-subsystem-1-localhost', res)

    def test__get_nvme_uuid(self):
        vol = mock.Mock()
        res = self.target._get_nvme_uuid(vol)
        self.assertEqual(vol.name_id, res)

    def test__get_nqns_for_location_no_subsystems(self):
        self.mock_object(self.target._nvmet_root, 'subsystems', iter([]))
        res = self.target._get_nqns_for_location(mock.sentinel.location)
        self.assertListEqual([], res)

    def test__get_nqns_for_location_no_subsystems_found(self):
        ns1 = mock.Mock(**{'get_attr.return_value': mock.sentinel.location1})
        subsys1 = mock.Mock(namespaces=iter([ns1]))

        ns2 = mock.Mock(**{'get_attr.return_value': mock.sentinel.location2})
        subsys2 = mock.Mock(namespaces=iter([ns2]))

        subsys = iter([subsys1, subsys2])
        self.mock_object(self.target._nvmet_root, 'subsystems', subsys)

        res = self.target._get_nqns_for_location(mock.sentinel.location3)

        self.assertListEqual([], res)
        ns1.get_attr.assert_called_once_with('device', 'path')
        ns2.get_attr.assert_called_once_with('device', 'path')

    def test__get_nqns_for_location_subsystems_found(self):
        ns1 = mock.Mock(**{'get_attr.return_value': mock.sentinel.location1})
        subsys1 = mock.Mock(namespaces=iter([ns1]))

        ns2 = mock.Mock(**{'get_attr.return_value': mock.sentinel.location2})
        ns1b = mock.Mock(**{'get_attr.return_value': mock.sentinel.location1})
        ns3 = mock.Mock(**{'get_attr.return_value': mock.sentinel.location3})
        subsys2 = mock.Mock(namespaces=iter([ns2, ns1b, ns3]))

        ns4 = mock.Mock(**{'get_attr.return_value': mock.sentinel.location4})
        subsys3 = mock.Mock(namespaces=iter([ns4]))

        subsys4 = mock.Mock(namespaces=iter([]))

        subsys = iter([subsys1, subsys2, subsys3, subsys4])
        self.mock_object(self.target._nvmet_root, 'subsystems', subsys)

        res = self.target._get_nqns_for_location(mock.sentinel.location1)

        self.assertListEqual([subsys1.nqn, subsys2.nqn, subsys4.nqn], res)
        ns1.get_attr.assert_called_once_with('device', 'path')
        ns2.get_attr.assert_called_once_with('device', 'path')
        ns1b.get_attr.assert_called_once_with('device', 'path')
        ns3.get_attr.assert_not_called()
        ns4.get_attr.assert_called_once_with('device', 'path')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils

from cinder.api import microversions as mv
from cinder.tests.functional import api_samples_test_base as test_base


@test_base.VolumesSampleBase.use_versions(
    mv.BASE_VERSION,  # 3.0
    mv.BACKUP_UPDATE,  # 3.9
    mv.BACKUP_PROJECT,  # 3.18
    mv.BACKUP_METADATA,  # 3.43
    mv.SUPPORT_COUNT_INFO,  # 3.45
    mv.BACKUP_PROJECT_USER_ID)  # 3.56
class BackupClassesSampleJsonTest(test_base.VolumesSampleBase):
    sample_dir = "backups"

    def setUp(self):
        super(BackupClassesSampleJsonTest, self).setUp()
        res = self._create_volume()
        res = jsonutils.loads(res.content)['volume']
        self._poll_volume_while(res['id'], ['creating'])
        self.subs = {
            "volume_id": res['id']
        }
        with self.common_api_sample():
            self.response = self._do_post('backups',
                                          'backup-create-request',
                                          self.subs)

    def test_backup_create(self):
        self._verify_response('backup-create-response',
                              {}, self.response, 202)

    @test_base.VolumesSampleBase.override_mv(mv.BASE_VERSION)  # 3.0
    def test_backup_list(self):
        response = self._do_get('backups')
        self._verify_response('backups-list-response',
                              {}, response, 200)

    @test_base.VolumesSampleBase.override_mv(mv.SUPPORT_COUNT_INFO)  # 3.45
    def test_backup_list_with_count(self):
        response = self._do_get('backups?with_count=True')
        self._verify_response('backups-list-response',
                              {}, response, 200)

    @test_base.VolumesSampleBase.override_mv(mv.BACKUP_UPDATE)  # 3.9
    def test_backup_update(self):
        res = jsonutils.loads(self.response.content)['backup']
        response = self._do_put('backups/%s' % res['id'],
                                'backup-update-request')
        self._verify_response('backup-update-response',
                              {}, response, 200)

    def test_backup_show(self):
        res = jsonutils.loads(self.response.content)['backup']
        response = self._do_get('backups/%s' % res['id'])
        self._verify_response('backup-show-response',
                              {}, response, 200)

    @test_base.VolumesSampleBase.override_mv(mv.BASE_VERSION)  # 3.0
    def test_backup_list_detail(self):
        response = self._do_get('backups/detail')
        self._verify_response('backups-list-detailed-response',
                              {}, response, 200)

    @test_base.VolumesSampleBase.override_mv(mv.SUPPORT_COUNT_INFO)  # 3.45
    def test_backup_list_detail_with_count(self):
        response = self._do_get('backups/detail?with_count=True')
        self._verify_response('backups-list-detailed-response',
                              {}, response, 200)



# Copyright 2023 toyou Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import base64
import json
import time

from oslo_log import log as logging
from oslo_utils import netutils
import requests

from cinder import exception
from cinder.i18n import _

LOG = logging.getLogger(__name__)


class TydsClient(object):
    def __init__(self, hostname, port, username, password):
        """Initializes a new instance of the TydsClient.

        :param hostname: IP address of the Toyou distributed storage system.
        :param port: The port to connect to the Toyou distributed storage
        system.
        :param username: The username for authentication.
        :param password: The password for authentication.
        """
        self._username = username
        self._password = base64.standard_b64encode(password.encode('utf-8')
                                                   ).decode('utf-8')
        self._baseurl = f"http://{hostname}:{port}/api"
        self._snapshot_count = 999
        self._token = None
        self._token_expiration = 0
        self._ip = self._get_local_ip()

    def get_token(self):
        if self._token and time.time() < self._token_expiration:
            # Token is not expired, directly return the existing Token
            return self._token

        # Token has expired or has not been obtained before,
        # retrieving the Token again
        self._token = self.login()
        self._token_expiration = time.time() + 710 * 60
        return self._token

    def send_http_api(self, url, params=None, method='post'):
        """Send an HTTP API request to the storage.

        :param url: The URL for the API request.
        :param params: The parameters for the API request.
        :param method: The HTTP method for the API request. Default is 'post'.

        :return: The response from the API request.

        :raises VolumeBackendAPIException: If the API request fails.

        """
        if params:
            params = json.dumps(params)

        url = f"{self._baseurl}/{url}"
        header = {
            'Authorization': self.get_token(),
            'Content-Type': 'application/json'
        }

        LOG.debug(
            "Toyou Cinder Driver Requests: http_process header: %(header)s "
            "url: %(url)s method: %(method)s",
            {'header': header, 'url': url, 'method': method}
        )

        response = self.do_request(method, url, header, params)
        return response

    @staticmethod
    def _get_local_ip():
        """Get the local IP address.

        :return: The local IP address.

        """
        return netutils.get_my_ipv4()

    def login(self):
        """Perform login to obtain an authentication token.

        :return: The authentication token.

        :raises VolumeBackendAPIException: If the login request fails or the
                                            authentication token cannot be
                                            obtained.

        """
        params = {
            'REMOTE_ADDR': self._ip,
            'username': self._username,
            'password': self._password
        }
        data = json.dumps(params)
        url = f"{self._baseurl}/auth/login/"
        response = self.do_request(method='post',
                                   url=url,
                                   header={'Content-Type': 'application/json'},
                                   data=data)
        self._token = response.get('token')
        return self._token

    @staticmethod
    def do_request(method, url, header, data):
        """Send request to the storage and handle the response.

        :param method: The HTTP method to use for the request. Valid methods
                       are 'post', 'get', 'put', and 'delete'.
        :param url: The URL to send the request to.
        :param header: The headers to include in the request.
        :param data: The data to send in the request body.

        :return: The response data returned by the storage system.

        :raises VolumeBackendAPIException: If the request fails or the response
                                            from the storage system is not as
                                            expected.

        """
        valid_methods = ['post', 'get', 'put', 'delete']
        if method not in valid_methods:
            raise exception.VolumeBackendAPIException(
                data=_('Unsupported request type: %s.') % method
            )

        try:
            req = getattr(requests, method)(url, data=data, headers=header)
            req.raise_for_status()
            response = req.json()
        except requests.exceptions.RequestException as e:
            msg = (_('Request to %(url)s failed: %(error)s') %
                   {'url': url, 'error': str(e)})
            raise exception.VolumeBackendAPIException(data=msg)
        except ValueError as e:
            msg = (_('Failed to parse response from %(url)s: %(error)s') %
                   {'url': url, 'error': str(e)})
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('URL: %(url)s, TYPE: %(type)s, CODE: %(code)s, '
                  'RESPONSE: %(response)s.',
                  {'url': req.url,
                   'type': method,
                   'code': req.status_code,
                   'response': response})
        # Response Error
        if response.get('code') != '0000':
            msg = (_('ERROR RESPONSE: %(response)s URL: %(url)s PARAMS: '
                     '%(params)s.') %
                   {'response': response, 'url': url, 'params': data})
            raise exception.VolumeBackendAPIException(data=msg)
        # return result
        return response.get('data')

    def get_pools(self):
        """Query pool information.

        :return: A list of pool information.

        """
        url = 'pool/pool/'
        response = self.send_http_api(url=url, method='get')
        pool_list = response.get('poolList', [])
        return pool_list

    def get_volumes(self):
        """Query volume information.

        :return: A list of volume information.

        """
        url = 'block/blocks'
        vol_list = self.send_http_api(url=url, method='get').get('blockList')
        return vol_list

    def create_volume(self, vol_name, size, pool_name, stripe_size):
        """Create a volume.

        :param vol_name: The name of the volume.
        :param size: The size of the volume in MB.
        :param pool_name: The name of the pool to create the volume in.
        :param stripe_size: The stripe size of the volume.
        :return: The response from the API call.

        """
        url = 'block/blocks/'
        params = {'blockName': vol_name,
                  'sizeMB': size,
                  'poolName': pool_name,
                  'stripSize': stripe_size}
        return self.send_http_api(url=url, method='post', params=params)

    def delete_volume(self, vol_id):
        """Delete a volume.

        :param vol_id: The ID of the volume to delete.

        """
        url = 'block/recycle/forceCreate/'
        params = {'id': [vol_id]}
        self.send_http_api(url=url, method='post', params=params)

    def extend_volume(self, vol_name, pool_name, size_mb):
        """Extend the size of a volume.

        :param vol_name: The name of the volume to extend.
        :param pool_name: The name of the pool where the volume resides.
        :param size_mb: The new size of the volume in MB.

        """
        url = 'block/blocks/%s/' % vol_name
        params = {'blockName': vol_name,
                  'sizeMB': size_mb,
                  'poolName': pool_name}
        self.send_http_api(url=url, method='put', params=params)

    def create_clone_volume(self, *args):
        """Create a clone of a volume.

        :param args: The arguments needed for cloning a volume.
            Args:
                - pool_name: The name of the source pool.
                - block_name: The name of the source block.
                - block_id: The ID of the source block.
                - target_pool_name: The name of the target pool.
                - target_pool_id: The ID of the target pool.
                - target_block_name: The name of the target block.

        """
        pool_name, block_name, block_id, target_pool_name, target_pool_id, \
            target_block_name = args
        params = {
            'poolName': pool_name,
            'blockName': block_name,
            'blockId': block_id,
            'copyType': 0,  # 0 means shallow copy, currently copy volume first
            # default shallow copy, 1 means deep copy
            'metapoolName': 'NULL',
            'targetMetapoolName': 'NULL',
            'targetPoolName': target_pool_name,
            'targetPoolId': target_pool_id,
            'targetBlockName': target_block_name
        }
        url = 'block/block/copy/'
        self.send_http_api(url=url, params=params)

    def get_snapshot(self, volume_id=None):
        """Get a list of snapshots.

        :param volume_id: The ID of the volume to filter snapshots (default:
                          None).
        :return: The list of snapshots.

        """
        url = 'block/snapshot?pageNumber=1'
        if volume_id:
            url += '&blockId=%s' % volume_id
        url += '&pageSize=%s'
        response = self.send_http_api(
            url=url % self._snapshot_count, method='get')
        if self._snapshot_count < int(response.get('total')):
            self._snapshot_count = int(response.get('total'))
            response = self.send_http_api(
                url=url % self._snapshot_count, method='get')
        snapshot_list = response.get('snapShotList')
        return snapshot_list

    def create_snapshot(self, name, volume_id, comment=''):
        """Create a snapshot of a volume.

        :param name: The name of the snapshot.
        :param volume_id: The ID of the volume to create a snapshot from.
        :param comment: The optional comment for the snapshot (default: '').

        """
        url = 'block/snapshot/'
        params = {'sourceBlock': volume_id,
                  'snapShotName': name,
                  'remark': comment}
        self.send_http_api(url=url, method='post', params=params)

    def delete_snapshot(self, snapshot_id):
        """Delete a snapshot.

        :param snapshot_id: The ID of the snapshot to delete.

        """
        url = 'block/snapshot/%s/' % snapshot_id
        self.send_http_api(url=url, method='delete')

    def create_volume_from_snapshot(self, volume_name, pool_name,
                                    snapshot_name, source_volume_name,
                                    source_pool_name):
        """Create a volume from a snapshot.

        :param volume_name: The name of the new volume.
        :param pool_name: The name of the pool for the new volume.
        :param snapshot_name: The name of the snapshot to create the volume
                              from.
        :param source_volume_name: The name of the source volume (snapshot's
                                   origin).
        :param source_pool_name: The name of the pool for the source volume.

        """
        url = 'block/clone/'
        params = {'cloneBlockName': volume_name,
                  'targetPoolName': pool_name,
                  'snapName': snapshot_name,
                  'blockName': source_volume_name,
                  'poolName': source_pool_name,
                  'targetMetapoolName': 'NULL'}
        self.send_http_api(url=url, method='post', params=params)

    def get_clone_progress(self, volume_id, volume_name):
        """Get the progress of a volume clone operation.

        :param volume_id: The ID of the volume being cloned.
        :param volume_name: The name of the volume being cloned.
        :return: The progress of the clone operation.

        """
        url = 'block/clone/progress/'
        params = {'blockId': volume_id,
                  'blockName': volume_name}
        progress = self.send_http_api(url=url, method='post', params=params)
        return progress

    def get_copy_progress(self, block_id, block_name, target_block_name):
        """Get the progress of a block copy operation.

        :param block_id: The ID of the block being copied.
        :param block_name: The name of the block being copied.
        :param target_block_name: The name of the target block.
        :return: The progress of the copy operation.

        """
        url = 'block/block/copyprogress/'
        params = {
            'blockId': block_id,
            'blockName': block_name,
            'targetBlockName': target_block_name
        }

        progress_data = self.send_http_api(url=url, params=params)
        return progress_data

    def create_initiator_group(self, group_name, client):
        """Create an initiator group.

        :param group_name: The name of the initiator group.
        :param client: The client information for the initiator group.
        """
        url = 'iscsi/client-group/'
        params = {
            'group_name': group_name,
            'client': client,
            'chap_auth': 0,
            'mode': 'ISCSI'
        }
        self.send_http_api(url=url, params=params)

    def delete_initiator_group(self, group_id):
        """Delete an initiator group.

        :param group_id: The ID of the initiator group.
        :return: The response from the API call.
        """
        url = 'iscsi/client-group/?group_id=%s' % group_id
        return self.send_http_api(url=url, method='delete')

    def get_initiator_list(self):
        """Get the list of initiators.

        :return: The list of initiators.
        """
        url = 'iscsi/client-group/'
        res = self.send_http_api(url=url, method='get')
        initiator_list = res.get('client_group_list')
        return initiator_list

    def get_target(self):
        """Get the list of target hosts.

        :return: The list of target hosts.
        """
        url = '/host/host/'
        res = self.send_http_api(url=url, method='get')
        target = res.get('hostList')
        return target

    def create_target(self, group_name, target_list, vols_info):
        """Create a target.

        :param group_name: The name of the initiator group.
        :param target_list: The list of target hosts.
        :param vols_info: The information of the volumes.
        :return: The response from the API call.
        """
        url = 'iscsi/target/'
        params = {"group_name": group_name,
                  "chap_auth": 0,
                  "write_cache": 1,
                  "hostName": ",".join(target_list),
                  "block": vols_info}
        return self.send_http_api(url=url, params=params, method='post')

    def delete_target(self, target_name):
        """Delete a target.

        :param target_name: The name of the target.
        :return: The response from the API call.
        """
        url = 'iscsi/target/?targetIqn=%s' % target_name
        return self.send_http_api(url=url, method='delete')

    def modify_target(self, target_name, target_list, vol_info):
        """Modify a target.

        :param target_name: The name of the target.
        :param target_list: The list of target hosts.
        :param vol_info: The information of the volumes.
        :return: The response from the API call.
        """
        url = 'iscsi/target/'
        params = {
            "targetIqn": target_name,
            "chap_auth": 0,
            "hostName": target_list,
            "block": vol_info
        }
        return self.send_http_api(url=url, params=params, method='put')

    def get_initiator_target_connections(self):
        """Get the list of IT (Initiator-Target) connections.

        :return: The list of IT connections.
        """
        url = 'iscsi/target/'
        res = self.send_http_api(url=url, method='get')
        target_list = res.get('target_list')
        return target_list

    def generate_config(self, target_name):
        """Generate configuration for a target.

        :param target_name: The name of the target.
        """
        url = 'iscsi/target-config/'
        params = {
            'targetName': target_name
        }
        self.send_http_api(url=url, params=params, method='post')

    def restart_service(self, host_name):
        """Restart the iSCSI service on a host.

        :param host_name: The name of the host.
        """
        url = 'iscsi/service/restart/'
        params = {
            "hostName": host_name
        }
        self.send_http_api(url=url, params=params, method='post')



# Copyright (c) 2017 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from cinder import context
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit.volume.drivers.dell_emc import powerflex
from cinder.tests.unit.volume.drivers.dell_emc.powerflex import mocks


class TestInitializeConnectionSnapshot(powerflex.TestPowerFlexDriver):

    def setUp(self):
        super(TestInitializeConnectionSnapshot, self).setUp()
        self.snapshot_id = 'SNAPID'
        self.ctx = context.RequestContext('fake', 'fake', auth_token=True)
        self.fake_path = '/fake/path/vol-xx'
        self.volume = fake_volume.fake_volume_obj(
            self.ctx, **{'provider_id': fake.PROVIDER_ID})
        self.connector = {'sdc_guid': 'fake_guid'}

        self.sdc = {
            "id": "sdc1",
        }
        self.HTTPS_MOCK_RESPONSES = {
            self.RESPONSE_MODE.Valid: {
                'types/Sdc/instances':
                [{'id': "sdc1", 'sdcGuid': 'fake_guid'}],
                'instances/Volume::{}/action/setMappedSdcLimits'.format(
                    self.snapshot_id
                ): mocks.MockHTTPSResponse({}, 200),
            },
        }

    def test_backup_can_use_snapshots(self):
        """Make sure the driver can use snapshots for backup."""
        use_snaps = self.driver.backup_use_temp_snapshot()
        self.assertTrue(use_snaps)

    def test_initialize_connection_without_size(self):
        """Test initializing when we do not know the snapshot size.

        ScaleIO can determine QOS specs based upon volume/snapshot size
        The QOS keys should not be returned
        """
        snapshot = fake_snapshot.fake_snapshot_obj(
            self.ctx, **{'volume': self.volume,
                         'provider_id': self.snapshot_id})
        self.driver._attach_volume_to_host = mock.MagicMock(
            return_value=None
        )
        self.driver._check_volume_mapped = mock.MagicMock(
            return_value=None
        )
        props = self.driver.initialize_connection_snapshot(
            snapshot,
            self.connector)
        # validate the volume type
        self.assertEqual(props['driver_volume_type'], 'scaleio')
        # make sure a volume name and id exist
        self.assertIsNotNone(props['data']['scaleIO_volname'])
        self.assertEqual(self.snapshot_id,
                         props['data']['scaleIO_volume_id'])
        # make sure QOS properties are not set
        self.assertNotIn('iopsLimit', props['data'])

    def test_initialize_connection_with_size(self):
        """Test initializing when we know the snapshot size.

        PowerFlex can determine QOS specs based upon volume/snapshot size
        The QOS keys should not be returned
        """
        snapshot = fake_snapshot.fake_snapshot_obj(
            self.ctx, **{'volume': self.volume,
                         'provider_id': self.snapshot_id,
                         'volume_size': 8})
        self.driver._attach_volume_to_host = mock.MagicMock(
            return_value=None
        )
        self.driver._check_volume_mapped = mock.MagicMock(
            return_value=None
        )
        props = self.driver.initialize_connection_snapshot(
            snapshot,
            self.connector)
        # validate the volume type
        self.assertEqual(props['driver_volume_type'], 'scaleio')
        # make sure a volume name and id exist
        self.assertIsNotNone(props['data']['scaleIO_volname'])
        self.assertEqual(self.snapshot_id,
                         props['data']['scaleIO_volume_id'])
        # make sure QOS properties are not set
        self.assertNotIn('iopsLimit', props['data'])

    def test_qos_specs(self):
        """Ensure QOS specs are honored if present."""
        qos = {'maxIOPS': 1000, 'maxBWS': 2048}
        snapshot = fake_snapshot.fake_snapshot_obj(
            self.ctx, **{'volume': self.volume,
                         'provider_id': self.snapshot_id,
                         'volume_size': 8})
        extraspecs = {}
        self.driver._get_volumetype_qos = mock.MagicMock()
        self.driver._get_volumetype_qos.return_value = qos
        self.driver._get_volumetype_extraspecs = mock.MagicMock()
        self.driver._get_volumetype_extraspecs.return_value = extraspecs
        self.driver._attach_volume_to_host = mock.MagicMock(
            return_value=None
        )
        self.driver._check_volume_mapped = mock.MagicMock(
            return_value=None
        )
        self.driver.primary_client.set_sdc_limits = mock.MagicMock()
        self.driver.initialize_connection_snapshot(
            snapshot,
            self.connector)

        self.driver.primary_client.set_sdc_limits.assert_called_once_with(
            self.snapshot_id, self.sdc["id"], '2048', '1000')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from oslo_log import log as logging
from oslo_utils import uuidutils

from cinder import exception
from cinder.privsep.targets import nvmet
from cinder import utils
from cinder.volume.targets import nvmeof


LOG = logging.getLogger(__name__)


class NVMETTargetAddError(exception.CinderException):
    message = "Failed to add subsystem: %(subsystem)s"


class NVMETTargetDeleteError(exception.CinderException):
    message = "Failed to delete subsystem: %(subsystem)s"


class NVMET(nvmeof.NVMeOF):
    SHARED_TARGET_SUPPORT = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._nvmet_root = nvmet.Root()

    # #######  Connection initiation methods ########

    def initialize_connection(self, volume, connector):
        """Create an export & map if shared."""
        # Non-shared connections was the original implementation where all the
        # export & mapping was done on export and the connection info was
        # stored in the volume, so let the original implementation handle it.
        if not self.share_targets:
            return super().initialize_connection(volume, connector)

        # For the shared case the export only stores the path of the volume
        volume_path = volume.provider_location
        if not os.path.exists(volume_path):
            raise exception.InvalidConfigurationValue(
                'Target driver configured with shared targets, but volume '
                'exported as non shared.')

        nqn, ns_id = self._map_volume(volume, volume_path, connector)
        uuid = self._get_nvme_uuid(volume)
        return {
            'driver_volume_type': self.protocol,
            'data': self._get_connection_properties(nqn,
                                                    self.target_ips,
                                                    self.target_port,
                                                    self.nvme_transport_type,
                                                    ns_id, uuid),
        }

    def create_export(self, context, volume, volume_path):
        """Create an export & map if not shared."""
        # For shared targets everything gets done on initialize_connection
        if self.share_targets:
            location = volume_path
        else:
            nqn, ns_id = self._map_volume(volume, volume_path)
            location = self.get_nvmeof_location(nqn,
                                                self.target_ips,
                                                self.target_port,
                                                self.nvme_transport_type,
                                                ns_id)

        return {'location': location, 'auth': ''}

    @utils.synchronized('nvmetcli', external=True)
    def _map_volume(self, volume, volume_path, connector=None):
        """Ensure a volume is exported and mapped in nvmet."""
        # Create NVME subsystem for previously created LV
        nqn = self._get_target_nqn(volume.id, connector)
        try:
            uuid = self._get_nvme_uuid(volume)

            ns_id = self._ensure_subsystem_exists(nqn, volume_path, uuid)

            self._ensure_port_exports(nqn, self.target_ips, self.target_port,
                                      self.nvme_transport_type,
                                      self.nvmet_port_id)
        except Exception:
            LOG.error('Failed to add subsystem: %s', nqn)
            raise NVMETTargetAddError(subsystem=nqn)

        LOG.info('Subsystem %s now exported on port %s', nqn, self.target_port)
        return nqn, ns_id

    def _ensure_subsystem_exists(self, nqn, volume_path, uuid):
        """Ensure a subsystem and namespace exist in nvmet."""
        # Assume if subsystem exists, it has the right configuration
        try:
            subsystem = nvmet.Subsystem(nqn)
            LOG.debug('Skip creating subsystem %s as it already exists.', nqn)

            ns_id = self._ensure_namespace_exists(subsystem, volume_path, uuid)
            return ns_id

        except nvmet.NotFound:
            LOG.debug('Creating subsystem %s.', nqn)

        ns_id = self.nvmet_ns_id
        subsystem_section = {
            "allowed_hosts": [],
            "attr": {
                "allow_any_host": "1"
            },
            "namespaces": [self._namespace_dict(uuid, volume_path, ns_id)],
            "nqn": nqn}

        nvmet.Subsystem.setup(subsystem_section)  # privsep
        LOG.debug('Added subsystem: %s', nqn)
        return ns_id

    def _namespace_dict(self, uuid, volume_path, ns_id):
        """Build the dict data for a new namespace in nvmet library format."""
        if self.share_targets:
            nguid = uuid
            LOG.debug('Sharing subsystem, using nguid = uuid = %s', nguid)
        else:
            nguid = str(uuidutils.generate_uuid())
            LOG.debug('Not sharing subsystem, using randmo nguid = %s', nguid)
        return {
            "device": {
                "nguid": nguid,
                "uuid": uuid,
                "path": volume_path,
            },
            "enable": 1,
            "nsid": ns_id
        }

    def _ensure_namespace_exists(self, subsystem, volume_path, uuid):
        """Ensure the namespace exists in nvmet."""
        for ns in subsystem.namespaces:
            if ns.get_attr('device', 'path') == volume_path:
                return ns.nsid

        ns_id = self._get_available_namespace_id(subsystem)
        ns_data = self._namespace_dict(uuid, volume_path, ns_id)
        nvmet.Namespace.setup(subsystem, ns_data)
        return ns_id

    def _get_available_namespace_id(self, subsystem):
        """Get the next available ns_id.

        Shared targets will have multiple namespaces under the same subsystem,
        so we cannot use self.nvmet_ns_id for them all.

        This method searches for an available namespace id in the provided
        subsystem considering all ids below self.nvmet_ns_id as reserved.

        We cannot let the nvmet library assign it automatically because it
        starts assigning from 1.

        For non shared the method returns configured nvmet_ns_id.
        """
        minimum = self.nvmet_ns_id

        if not self.share_targets:
            return minimum

        used = [ns.nsid for ns in subsystem.namespaces if ns.nsid >= minimum]

        if not used:
            return minimum

        higher = max(used)
        # If there are no gaps return the next available id
        if len(used) > higher - minimum:
            if higher == nvmet.Namespace.MAX_NSID:
                raise Exception('Reached max namespaces in subsystem')
            return higher + 1

        # Find an id in the gaps.  Don't include higher, as we know it's used
        available = set(range(minimum, higher)).difference(used)
        return available.pop()

    def _get_nvme_uuid(self, volume):
        return volume.name_id

    def _ensure_port_exports(self, nqn, addrs, port, transport_type, port_id):
        for addr in addrs:
            # Assume if port exists, it has the right configuration
            try:
                nvme_port = nvmet.Port(port_id)
                LOG.debug('Skip creating port %s as it already exists.',
                          port_id)
            except nvmet.NotFound:
                LOG.debug('Creating port %s.', port_id)

                # Port section
                port_section = {
                    "addr": {
                        "adrfam": "ipv4",
                        "traddr": addr,
                        "treq": "not specified",
                        "trsvcid": port,
                        "trtype": transport_type,
                    },
                    "portid": port_id,
                    "referrals": [],
                    "subsystems": [nqn]
                }
                nvmet.Port.setup(self._nvmet_root, port_section)  # privsep
                LOG.debug('Added port: %s', port_id)

            else:
                if nqn in nvme_port.subsystems:
                    LOG.debug('%s already exported on port %s', nqn, port_id)
                else:
                    nvme_port.add_subsystem(nqn)  # privsep
                    LOG.debug('Exported %s on port %s', nqn, port_id)
            port_id += 1

    # #######  Connection termination methods ########

    def terminate_connection(self, volume, connector, **kwargs):
        """Remove the mapping for shared."""
        # TODO: Add support for force and other parameters
        if self.share_targets:
            self._locked_unmap_volume(volume, connector)
            LOG.info('Volume %s is no longer exported', volume.id)

    def remove_export(self, context, volume):
        """Remove the mapping for non shared."""
        if not self.share_targets:
            self._locked_unmap_volume(volume)
            LOG.info('Volume %s is no longer exported', volume.id)

    @utils.synchronized('nvmetcli', external=True)
    def _locked_unmap_volume(self, volume, connector=None):
        """Remove volume's ns from subsystem and subsystem if empty."""
        if connector or not self.share_targets:
            nqns = [self._get_target_nqn(volume.id, connector)]
        else:
            # We need to remove all existing maps (we are sharing)
            LOG.debug('Removing EVERYTHING for volume %s', volume.id)
            nqns = self._get_nqns_for_location(volume.provider_location)

        exceptions = []
        for nqn in nqns:
            try:
                self._unmap_volume(volume, nqn)
            except Exception as exc:
                exceptions.append(exc)

        # TODO: Once we only support Python 3.11+ use ExceptionGroup to raise
        # all the exceptions.
        if exceptions:
            raise exceptions[0]

    def _unmap_volume(self, volume, nqn):
        try:
            subsystem = nvmet.Subsystem(nqn)
        except nvmet.NotFound:
            LOG.info('Skipping unmapping. No NVMe subsystem for volume: %s',
                     volume.id)
            return

        if self.share_targets:
            volume_path = volume.provider_location
            for ns in subsystem.namespaces:
                if ns.get_attr('device', 'path') == volume_path:
                    LOG.debug('Deleting namespace %s', ns.nsid)
                    ns.delete()  # privsep call
                    break

            # If there are still namespaces we cannot remove the subsystem
            if any(s for s in subsystem.namespaces):
                return

        for port in self._nvmet_root.ports:
            if nqn in port.subsystems:
                LOG.debug('Removing %s from port %s', nqn, port.portid)
                port.remove_subsystem(nqn)  # privsep call

        LOG.debug('Deleting %s', nqn)
        subsystem.delete()  # privsep call
        LOG.info('Subsystem %s removed', nqn)

    # #######  General methods ########

    def _get_target_nqn(self, volume_id, connector):
        # For shared targets the subsystem is named after the host
        if self.share_targets:
            postfix = connector['host']
        else:
            postfix = volume_id
        return f'nqn.{self.nvmet_subsystem_name}-{postfix}'

    def _get_nqns_for_location(self, provider_location):
        """Get all subystem nqns for a give provider location.

        This also returns empty subsystems, since we don't know if those were
        created to try to use them for the volume of the provider_location and
        failed during the creation.

        This method needs to be called within the nvmetcli locked section.
        """
        nqns = []
        for subsys in self._nvmet_root.subsystems:
            empty = True  # subsytems is an iterable, can check it with bool
            found = False
            for ns in subsys.namespaces:
                empty = False
                if ns.get_attr('device', 'path') == provider_location:
                    found = True
                    break
            if found or empty:
                nqns.append(subsys.nqn)
        return nqns



#    Copyright (c) 2020 Open-E, Inc.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

jdss_connection_opts = [
    cfg.ListOpt('san_hosts',
                default='',
                help='IP address of Open-E JovianDSS SA'),
    cfg.IntOpt('jovian_recovery_delay',
               default=60,
               help='Time before HA cluster failure.'),
    cfg.ListOpt('jovian_ignore_tpath',
                default=[],
                help='List of multipath ip addresses to ignore.'),
]

jdss_iscsi_opts = [
    cfg.IntOpt('chap_password_len',
               default=12,
               help='Length of the random string for CHAP password.'),
    cfg.StrOpt('jovian_pool',
               default='Pool-0',
               help='JovianDSS pool that holds all cinder volumes'),
]

jdss_volume_opts = [
    cfg.StrOpt('jovian_block_size',
               default='64K',
               choices=[('16K', 'Use 16K block size'),
                        ('32K', 'Use 32K block size'),
                        ('64K', 'Use 64K block size'),
                        ('128K', 'Use 128K block size'),
                        ('256K', 'Use 256K block size'),
                        ('512K', 'Use 512K block size'),
                        ('1M', 'Use 1M block size')],
               help='Block size for new volume')
]

CONF = cfg.CONF
CONF.register_opts(jdss_connection_opts)
CONF.register_opts(jdss_iscsi_opts)
CONF.register_opts(jdss_volume_opts)



#    Copyright (c) 2023 Open-E, Inc.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from cinder import exception
from cinder.tests.unit import test
from cinder.volume.drivers.open_e.jovian_common import jdss_common as jcom

UUID_1 = '12345678-1234-1234-1234-000000000001'
UUID_2 = '12345678-1234-1234-1234-000000000002'
UUID_S1 = '12345678-1234-1234-1234-100000000001'
UUID_S2 = '12345678-1234-1234-1234-100000000002'

V_UUID_1 = 'v_12345678-1234-1234-1234-000000000001'
V_UUID_2 = 'v_12345678-1234-1234-1234-000000000002'
V_UUID_3 = 'v_12345678-1234-1234-1234-000000000003'
S_UUID_1 = f's_{UUID_S1}_{UUID_1}'
S_UUID_2 = f's_{UUID_S2}_{UUID_1}'
T_UUID_1 = 't_12345678-1234-1234-1234-000000000001'


VOLUME_GET_THAT_IS_CLONE = {
    "origin": f"Pool-0/{V_UUID_1}@{S_UUID_1}",
    "relatime": None,
    "acltype": None,
    "vscan": None,
    "full_name": f"Pool-0/{jcom.vname(UUID_2)}",
    "userrefs": None,
    "primarycache": "all",
    "logbias": "latency",
    "creation": "1695078560",
    "sync": "always",
    "is_clone": True,
    "dedup": "off",
    "sharenfs": None,
    "receive_resume_token": None,
    "volsize": "1073741824",
    "referenced": "57344",
    "sharesmb": None,
    "createtxg": "19812058",
    "reservation": "0",
    "scontext": None,
    "mountpoint": None,
    "casesensitivity": None,
    "guid": "4947994863040470005",
    "usedbyrefreservation": "0",
    "dnodesize": None,
    "written": "0",
    "logicalused": "0",
    "compressratio": "1.00",
    "rootcontext": "none",
    "default_scsi_id": "5c02d042ed8dbce2",
    "type": "volume",
    "compression": "lz4",
    "snapdir": None,
    "overlay": None,
    "encryption": "off",
    "xattr": None,
    "volmode": "default",
    "copies": "1",
    "snapshot_limit": "18446744073709551615",
    "aclinherit": None,
    "defcontext": "none",
    "readonly": "off",
    "version": None,
    "recordsize": None,
    "filesystem_limit": None,
    "mounted": None,
    "mlslabel": "none",
    "secondarycache": "all",
    "refreservation": "0",
    "available": "954751713280",
    "san:volume_id": "5c02d042ed8dbce2570c8d5dc276dd6a2431e138",
    "encryptionroot": None,
    "exec": None,
    "refquota": None,
    "refcompressratio": "1.00",
    "quota": None,
    "utf8only": None,
    "keylocation": "none",
    "snapdev": "hidden",
    "snapshot_count": "18446744073709551615",
    "fscontext": "none",
    "clones": None,
    "canmount": None,
    "keystatus": None,
    "atime": None,
    "usedbysnapshots": "0",
    "normalization": None,
    "usedbychildren": "0",
    "volblocksize": "65536",
    "usedbydataset": "0",
    "objsetid": "19228",
    "name": "a2",
    "defer_destroy": None,
    "pbkdf2iters": "0",
    "checksum": "on",
    "redundant_metadata": "all",
    "filesystem_count": None,
    "devices": None,
    "keyformat": "none",
    "setuid": None,
    "used": "0",
    "logicalreferenced": "28672",
    "context": "none",
    "zoned": None,
    "nbmand": None,
}

SNAPSHOT_GET = {
    'referenced': '57344',
    'userrefs': '0',
    'primarycache': 'all',
    'creation': '2023-06-28 16:49:33',
    'volsize': '1073741824',
    'createtxg': '18402390',
    'guid': '15554334551928551694',
    'compressratio': '1.00',
    'rootcontext': 'none',
    'encryption': 'off',
    'defcontext': 'none',
    'written': '0',
    'type': 'snapshot',
    'secondarycache': 'all',
    'used': '0',
    'refcompressratio': '1.00',
    'fscontext': 'none',
    'objsetid': '106843',
    'name': S_UUID_1,
    'defer_destroy': 'off',
    'san:volume_id': 'e82c7fcbd78df0ffe67d363412e5091421d313ca',
    'mlslabel': 'none',
    'logicalreferenced': '28672',
    'context': 'none'}

SNAPSHOT_MULTIPLE_CLONES = {
    'referenced': '57344',
    'userrefs': '0',
    'primarycache': 'all',
    'creation': '2023-06-28 18:44:49',
    'volsize': '1073741824',
    'createtxg': '18403768',
    'guid': '18319280142829358721',
    'compressratio': '1.00',
    'rootcontext': 'none',
    'encryption': 'off',
    'defcontext': 'none',
    'written': '0',
    'type': 'snapshot',
    'secondarycache': 'all',
    'used': '0',
    'refcompressratio': '1.00',
    'fscontext': 'none',
    'objsetid': '107416',
    'clones': f'Pool-0/{V_UUID_2},Pool-0/{V_UUID_3}',
    'name': S_UUID_1,
    'defer_destroy': 'off',
    'san:volume_id': 'e82c7fcbd78df0ffe67d363412e5091421d313ca',
    'mlslabel': 'none',
    'logicalreferenced': '28672',
    'context': 'none'}

SNAPSHOTS_GET_NO_CLONES = [
    {'referenced': '57344',
     'userrefs': '0',
     'primarycache': 'all',
     'creation': '2023-06-28 16:49:33',
     'volsize': '1073741824',
     'createtxg': '18402390',
     'guid': '15554334551928551694',
     'compressratio': '1.00',
     'rootcontext': 'none',
     'encryption': 'off',
     'defcontext': 'none',
     'written': '0',
     'type': 'snapshot',
     'secondarycache': 'all',
     'used': '0',
     'refcompressratio': '1.00',
     'fscontext': 'none',
     'objsetid': '106843',
     'name': S_UUID_1,
     'defer_destroy': 'off',
     'san:volume_id': 'e82c7fcbd78df0ffe67d363412e5091421d313ca',
     'mlslabel': 'none',
     'logicalreferenced': '28672',
     'context': 'none'},
    {'referenced': '57344',
     'userrefs': '0',
     'primarycache': 'all',
     'creation': '2023-06-28 18:44:49',
     'volsize': '1073741824',
     'createtxg': '18403768',
     'guid': '18319280142829358721',
     'compressratio': '1.00',
     'rootcontext': 'none',
     'encryption': 'off',
     'defcontext': 'none',
     'written': '0',
     'type': 'snapshot',
     'secondarycache': 'all',
     'used': '0',
     'refcompressratio': '1.00',
     'fscontext': 'none',
     'objsetid': '107416',
     'name': S_UUID_2,
     'defer_destroy': 'off',
     'san:volume_id': 'e82c7fcbd78df0ffe67d363412e5091421d313ca',
     'mlslabel': 'none',
     'logicalreferenced': '28672',
     'context': 'none'}]


class TestOpenEJovianDSSCommon(test.TestCase):

    def test_is_volume(self):

        self.assertFalse(jcom.is_volume("asdasd"))
        self.assertFalse(jcom.is_volume(UUID_1))
        self.assertTrue(jcom.is_volume(V_UUID_1))

    def test_is_snapshot(self):

        self.assertFalse(jcom.is_snapshot("asdasd"))
        self.assertFalse(jcom.is_snapshot(UUID_S1))
        self.assertTrue(jcom.is_snapshot(S_UUID_1))

    def test_idname(self):

        self.assertEqual(UUID_1, jcom.idname(V_UUID_1))
        self.assertEqual(UUID_S1, jcom.idname(S_UUID_1))
        self.assertEqual(UUID_1, jcom.idname(T_UUID_1))

        self.assertRaises(exception.VolumeDriverException, jcom.idname, 'asd')

    def test_vname(self):

        self.assertEqual(V_UUID_1, jcom.vname(UUID_1))
        self.assertEqual(V_UUID_1, jcom.vname(V_UUID_1))
        self.assertRaises(exception.VolumeDriverException,
                          jcom.vname, S_UUID_1)

    def test_sname_to_id(self):

        self.assertEqual((UUID_S1, UUID_1), jcom.sname_to_id(S_UUID_1))

    def test_sid_from_sname(self):

        self.assertEqual(UUID_S1, jcom.sid_from_sname(S_UUID_1))

    def test_vid_from_sname(self):
        self.assertEqual(UUID_1, jcom.vid_from_sname(S_UUID_1))

    def test_sname(self):
        self.assertEqual(S_UUID_1, jcom.sname(UUID_S1, UUID_1))

    def test_sname_from_snap(self):

        snap = copy.deepcopy(SNAPSHOT_GET)
        self.assertEqual(S_UUID_1, jcom.sname_from_snap(snap))

    def test_is_hidden(self):

        self.assertTrue(jcom.is_hidden(T_UUID_1))
        self.assertFalse(jcom.is_hidden(S_UUID_1))

    def test_origin_snapshot(self):

        vol = copy.deepcopy(VOLUME_GET_THAT_IS_CLONE)

        self.assertEqual(S_UUID_1, jcom.origin_snapshot(vol))

    def test_origin_volume(self):

        vol = copy.deepcopy(VOLUME_GET_THAT_IS_CLONE)

        self.assertEqual(V_UUID_1, jcom.origin_volume(vol))

    def test_snapshot_clones(self):

        clones = [V_UUID_2, V_UUID_3]
        snap = copy.deepcopy(SNAPSHOT_MULTIPLE_CLONES)
        self.assertEqual(clones, jcom.snapshot_clones(snap))

    def test_hidden(self):

        self.assertEqual(T_UUID_1, jcom.hidden(V_UUID_1))

    def test_get_newest_snapshot_name(self):
        snaps = copy.deepcopy(SNAPSHOTS_GET_NO_CLONES)

        self.assertEqual(S_UUID_2, jcom.get_newest_snapshot_name(snaps))
